#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, Optional, Sequence, Union

import tensorflow as tf
from tensorflow.keras import backend
from tensorflow.python.keras.utils import control_flow_util
from tensorflow.python.ops import array_ops

from merlin.models.config.schema import requires_schema
from merlin.models.tf.blocks.core.base import Block, PredictionOutput
from merlin.models.tf.blocks.core.combinators import TabularBlock
from merlin.models.tf.typing import TabularData, TensorOrTabularData
from merlin.models.tf.utils.tf_utils import (
    df_to_tensor,
    get_candidate_probs,
    transform_label_to_onehot,
)
from merlin.models.utils import schema_utils
from merlin.schema import Schema, Tags


@Block.registry.register("as-sparse")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AsSparseFeatures(TabularBlock):
    """
    Convert inputs to sparse tensors.
    """

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs = {}
        for name, val in inputs.items():
            if isinstance(val, tuple):
                values = val[0][:, 0]
                row_lengths = val[1][:, 0]

                if values.dtype.is_floating:
                    values = tf.cast(values, tf.int32)
                if row_lengths.dtype.is_floating:
                    row_lengths = tf.cast(row_lengths, tf.int32)

                outputs[name] = tf.RaggedTensor.from_row_lengths(values, row_lengths).to_sparse()
            else:
                outputs[name] = val

        return outputs

    def compute_output_shape(self, input_shape):
        return input_shape


@Block.registry.register("as-dense")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AsDenseFeatures(TabularBlock):
    """Convert sparse inputs to dense tensors

    Parameters
    ----------
    max_seq_length : int
        The maximum length of multi-hot features.
    """

    def __init__(self, max_seq_length: Optional[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.max_seq_length = max_seq_length

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs = {}
        for name, val in inputs.items():
            if isinstance(val, tuple):
                values = val[0][:, 0]
                row_lengths = val[1][:, 0]
                ragged = tf.RaggedTensor.from_row_lengths(values, row_lengths)
                if self.max_seq_length:
                    outputs[name] = ragged.to_tensor(shape=[None, self.max_seq_length])
                else:
                    outputs[name] = tf.squeeze(ragged.to_tensor())
            else:
                outputs[name] = tf.squeeze(val)

        return outputs

    def compute_output_shape(self, input_shape):
        batch_size = self.calculate_batch_size_from_input_shapes(input_shape)
        outputs = {}

        for key, val in input_shape.items():
            if self.max_seq_length:
                outputs[key] = tf.TensorShape((batch_size, self.max_seq_length))
            else:
                # TODO: What to do here?
                raise ValueError("max_seq_length must be specified")

        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({"max_seq_length": self.max_seq_length})

        return config


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class RenameFeatures(TabularBlock):
    """Rename input features

    Parameters
    ----------
    renames: dict
        Mapping with new features names.
    schema: Schema, optional
        The `Schema` with input features,
        by default None
    """

    def __init__(
        self, renames: Dict[Union[str, Tags], str], schema: Optional[Schema] = None, **kwargs
    ):
        super().__init__(schema=schema, **kwargs)
        self.renames = {}
        for key, val in renames.items():
            if isinstance(key, Tags):
                if schema is None:
                    raise ValueError("Schema must be provided to rename features with Tags")
                cols = schema.select_by_tag(key)
                if len(cols) != 1:
                    raise ValueError(f"Tag: {key} does not uniquely identify a column")
                self.renames[cols.first.name] = val
            else:
                self.renames[key] = val

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs = {}

        for key, val in inputs.items():
            if key in self.renames:
                outputs[self.renames[key]] = val
            else:
                outputs[key] = val

        return outputs

    def compute_output_shape(self, input_shape):
        outputs = {}

        for key, val in input_shape.items():
            if key in self.renames:
                outputs[self.renames[key]] = val
            else:
                outputs[key] = val

        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({"renames": self.renames})

        return config


@Block.registry.register_with_multiple_names("stochastic-swap-noise", "ssn")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class StochasticSwapNoise(TabularBlock):
    """
    Applies Stochastic replacement of sequence features
    """

    def __init__(self, schema=None, pad_token=0, replacement_prob=0.1, **kwargs):
        super().__init__(**kwargs)
        self.schema = schema
        self.pad_token = pad_token
        self.replacement_prob = replacement_prob

    def call(
        self,
        inputs: TensorOrTabularData,
        input_mask: Optional[tf.Tensor] = None,
        training=False,
        **kwargs,
    ) -> TensorOrTabularData:
        def augment(input_mask):
            if self._schema:
                input_mask = input_mask or self.get_padding_mask_from_item_id(
                    inputs, self.pad_token
                )

            if isinstance(inputs, dict):
                return {key: self.augment(val, input_mask) for key, val in inputs.items()}

            return self.augment(inputs, input_mask)

        output = control_flow_util.smart_cond(training, lambda: augment(input_mask), lambda: inputs)

        return output

    def augment(self, input_tensor: tf.Tensor, mask: Optional[tf.Tensor], **kwargs) -> tf.Tensor:
        if mask is not None:
            if len(input_tensor.shape) == len(mask.shape) - 1:
                mask = mask[:, 0]

        casted = tf.cast(
            backend.random_binomial(array_ops.shape(input_tensor), p=self.replacement_prob),
            tf.int32,
        )

        replacement_mask_matrix = casted * tf.cast(mask, tf.int32)

        n_values_to_replace = tf.reduce_sum(replacement_mask_matrix)

        input_flattened_non_zero = tf.boolean_mask(
            input_tensor, tf.cast(replacement_mask_matrix, tf.bool)
        )

        sampled_values_to_replace = tf.gather(
            input_flattened_non_zero,
            tf.random.shuffle(tf.range(tf.shape(input_flattened_non_zero)[0]))[
                :n_values_to_replace
            ],
        )

        replacement_indices = tf.sparse.from_dense(replacement_mask_matrix).indices

        output_tensor = tf.tensor_scatter_nd_update(
            input_tensor, replacement_indices, sampled_values_to_replace
        )

        return output_tensor

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()

        config["pad_token"] = self.pad_token
        config["replacement_prob"] = self.replacement_prob

        return config


@Block.registry.register_with_multiple_names("continuous-powers")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ContinuousPowers(TabularBlock):
    """Trick from `Deep Neural Networks for YouTube Recommendations`"""

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs: TabularData = {}

        for key, val in inputs.items():
            outputs[key] = val
            if len(val.shape) < 2 or (len(val.shape) == 2 and val.shape[1] == 1):
                val_float = tf.cast(val, tf.float32)
                outputs[f"{key}_sqrt"] = tf.sqrt(val_float)
                outputs[f"{key}_pow"] = tf.pow(val_float, 2)

        return outputs

    def compute_output_shape(self, input_shape):
        output_shape = {}

        for key, val in input_shape.items():
            output_shape[key] = val
            if len(val) < 2 or (len(val) == 2 and val[1] == 1):
                output_shape[f"{key}_sqrt"] = val
                output_shape[f"{key}_squared"] = val

        return output_shape


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ExpandDims(TabularBlock):
    """
    Expand dims of selected input tensors.
    Example::

        inputs = {
            "cont_feat1": tf.random.uniform((NUM_ROWS,)),
            "cont_feat2": tf.random.uniform((NUM_ROWS,)),
            "multi_hot_categ_feat": tf.random.uniform(
                (NUM_ROWS, 4), minval=1, maxval=100, dtype=tf.int32
            ),
        }

        expand_dims_op = tr.ExpandDims(expand_dims={"cont_feat2": 0, "multi_hot_categ_feat": 1})
        expanded_inputs = expand_dims_op(inputs)
    """

    def __init__(self, expand_dims: Union[int, Dict[str, int]] = -1, **kwargs):
        """Instantiates the `ExpandDims` transformation, which allows to expand dims
        of the input tensors

        Parameters
        ----------
        expand_dims : Union[int, Dict[str, int]], optional, by default -1
            Defines which dimensions should be expanded. If an `int` is provided, all input tensors
            will have the same dimension expanded. If a `dict` is passed, only features matching
            the dict keys will be expanded, in the dimension specified as the dict values.
        """
        super().__init__(**kwargs)
        self.inputs_expand_dims = expand_dims

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs = {}

        for k, v in inputs.items():
            if isinstance(self.inputs_expand_dims, int):
                outputs[k] = tf.expand_dims(v, self.inputs_expand_dims)
            elif isinstance(self.inputs_expand_dims, dict) and k in self.inputs_expand_dims:
                expand_dim = self.inputs_expand_dims[k]
                outputs[k] = tf.expand_dims(v, expand_dim)
            elif self.inputs_expand_dims:
                outputs[k] = v
            else:
                raise ValueError("The expand_dims argument is not valid")

        return outputs

    def compute_output_shape(self, input_shape):
        return input_shape


@Block.registry.register_with_multiple_names("l2-norm")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class L2Norm(TabularBlock):
    """Apply L2-normalization to input tensors along a given axis"""

    def __init__(self, **kwargs):
        super(L2Norm, self).__init__(**kwargs)

    def call(self, inputs: Union[tf.Tensor, TabularData], axis: int = -1, **kwargs):
        if isinstance(inputs, dict):
            inputs = {key: tf.linalg.l2_normalize(inp, axis=axis) for key, inp in inputs.items()}
        else:
            inputs = tf.linalg.l2_normalize(inputs, axis=axis)

        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape


@Block.registry.register_with_multiple_names("remove_pad_3d")
@tf.keras.utils.register_keras_serializable(package="merlin_models")
class RemovePad3D(Block):
    """
    Flatten the sequence of labels and filter out non-targets positions

    Parameters
    ----------
        padding_idx: int
            The padding index value.
            Defaults to 0.
    Returns
    -------
        targets: tf.Tensor
            The flattened vector of true targets positions
        flatten_predictions: tf.Tensor
            If the predictions are 3-D vectors (sequential task),
            flatten the predictions vectors to keep only the ones related to target positions.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.padding_idx = 0

    def compute_output_shape(self, input_shape):
        return input_shape

    def call_outputs(
        self, outputs: PredictionOutput, training=True, **kwargs
    ) -> "PredictionOutput":
        targets, predictions = outputs.targets, outputs.predictions
        targets = tf.reshape(targets, (-1,))
        non_pad_mask = targets != self.padding_idx
        targets = tf.boolean_mask(targets, non_pad_mask)

        assert isinstance(predictions, tf.Tensor), "Predictions must be a tensor"

        if len(tuple(predictions.get_shape())) == 3:
            predictions = tf.reshape(predictions, (-1, predictions.shape[-1]))
            predictions = tf.boolean_mask(
                predictions, tf.broadcast_to(tf.expand_dims(non_pad_mask, 1), tf.shape(predictions))
            )

        return outputs.copy_with_updates(
            predictions=predictions,
            targets=targets,
        )


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class LogitsTemperatureScaler(Block):
    """Scale the logits higher or lower,
    this is often used to reduce the overconfidence of the model.

    Parameters
    ----------
    temperature : float
        Divide the logits by this scaler.
    apply_on_call_outputs: bool
        Whether to apply the transform (logits / temperature) on
        `call()` or `call_outputs()`. By default True
    """

    def __init__(self, temperature: float, apply_on_call_outputs: bool = True, **kwargs):
        super(LogitsTemperatureScaler, self).__init__(**kwargs)
        self.temperature = temperature
        self.apply_on_call_outputs = apply_on_call_outputs

    def call(self, inputs, training=False, **kwargs) -> tf.Tensor:
        if not self.apply_on_call_outputs:
            return self.apply_temperature(inputs)
        else:
            return inputs

    def call_outputs(
        self, outputs: PredictionOutput, training=True, **kwargs
    ) -> "PredictionOutput":
        targets, predictions = outputs.targets, outputs.predictions
        predictions = self.apply_temperature(predictions)

        return outputs.copy_with_updates(predictions=predictions, targets=targets)

    def compute_output_shape(self, input_shape):
        return input_shape

    def apply_temperature(self, predictions):
        assert isinstance(predictions, tf.Tensor), "Predictions must be a tensor"
        predictions = predictions / self.temperature
        return predictions


@Block.registry.register_with_multiple_names("weight-tying")
@tf.keras.utils.register_keras_serializable(package="merlin_models")
class ItemsPredictionWeightTying(Block):
    """Tying the item embedding weights with the output projection layer matrix [1]
    The output logits are obtained by multiplying the output vector by the item-ids embeddings.

    Parameters
    ----------
        schema : Schema
            The `Schema` with the input features
        bias_initializer : str, optional
            Initializer to use on the bias vector, by default "zeros"

    References:
    -----------
    [1] Hakan, Inan et al.
        "Tying word vectors and word classifiers: A loss framework for language modeling"
        arXiv:1611.01462
    """

    def __init__(self, schema: Schema, bias_initializer="zeros", **kwargs):
        super(ItemsPredictionWeightTying, self).__init__(**kwargs)
        self.bias_initializer = bias_initializer
        self.item_id_feature_name = schema.select_by_tag(Tags.ITEM_ID).column_names[0]
        self.num_classes = schema_utils.categorical_cardinalities(schema)[self.item_id_feature_name]
        self.item_domain = schema_utils.categorical_domains(schema)[self.item_id_feature_name]

    def build(self, input_shape):
        self.bias = self.add_weight(
            name="output_layer_bias",
            shape=(self.num_classes,),
            initializer=self.bias_initializer,
        )
        return super().build(input_shape)

    def call(self, inputs, training=False, **kwargs) -> tf.Tensor:
        embedding_table = self.context.get_embedding(self.item_domain)
        logits = tf.matmul(inputs, embedding_table, transpose_b=True)
        logits = tf.nn.bias_add(logits, self.bias)

        return logits


@Block.registry.register_with_multiple_names("categorical_to_onehot")
@tf.keras.utils.register_keras_serializable(package="merlin_models")
@requires_schema
class CategoricalOneHot(Block):
    """
    Transform categorical features (2-D and 3-D tensors) to a one-hot representation.

    Parameters:
    ----------
    schema : Optional[Schema]
        The `Schema` with the input features
    """

    def __init__(self, schema: Schema = None, **kwargs):
        super().__init__(**kwargs)
        if schema:
            self.set_schema(schema)
        self.cardinalities = schema_utils.categorical_cardinalities(self.schema)

    def build(self, input_shapes):
        super(CategoricalOneHot, self).build(input_shapes)

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        outputs = {}
        for name, val in self.cardinalities.items():
            outputs[name] = tf.squeeze(tf.one_hot(inputs[name], val))
        return outputs

    def compute_output_shape(self, input_shape):
        outputs = {}
        for key, val in input_shape.items():
            if len(val) == 3:
                outputs[key] = tf.TensorShape((val[0], val[1], self.cardinalities[key]))
            else:
                outputs[key] = tf.TensorShape((val[0], self.cardinalities[key]))
        return outputs

    def get_config(self):
        config = super().get_config()
        if self.schema:
            config["schema"] = schema_utils.schema_to_tensorflow_metadata_json(self.schema)
        return config


@Block.registry.register_with_multiple_names("label_to_onehot")
@tf.keras.utils.register_keras_serializable(package="merlin_models")
class LabelToOneHot(Block):
    """Transform the categorical encoded labels into a one-hot representation"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return input_shape

    def call_outputs(
        self, outputs: PredictionOutput, training=True, **kwargs
    ) -> "PredictionOutput":
        targets, predictions = outputs.targets, outputs.predictions

        num_classes = tf.shape(predictions)[-1]
        targets = transform_label_to_onehot(targets, num_classes)

        return outputs.copy_with_updates(targets=targets)


@tf.keras.utils.register_keras_serializable(package="merlin_models")
@requires_schema
class PopularityLogitsCorrection(Block):
    """Correct the predicted logit scores based on the item frequency,
    using the logQ correction proposed in sampled softmax [1]_ [2]_.
    The correction is done as `logits -= log(item_prob)`,
    where `item_prob = item_freq_count / sum(item_freq_count)` is
    a probability distribution of the item frequency. In a nutshell,
    the logQ correction aims to increase the prediction scores (logits)
    for infrequent items and decrease the ones for frequent items.

    References
    ----------
    .. [1] Yoshua Bengio and Jean-Sébastien Sénécal. 2003. Quick Training of Probabilistic
       Neural Nets by Importance Sampling. In Proceedings of the conference on Artificial
       Intelligence and Statistics (AISTATS).

    .. [2] Y. Bengio and J. S. Senecal. 2008. Adaptive Importance Sampling to Accelerate
       Training of a Neural Probabilistic Language Model. Trans. Neur. Netw. 19, 4 (April
       2008), 713–722. https://doi.org/10.1109/TNN.2007.912312

    Parameters:
    ----------
    item_freq_probs : Union[tf.Tensor, Sequence]
        A Tensor or list with item frequencies (if is_prob_distribution=False)
        or with item probabilities (if is_prob_distribution=True)
    is_prob_distribution: bool, optional
        If True, the item_freq_probs should be a probability distribution of the items.
        If False, the item frequencies is converted to probabilities
    reg_factor: float
        Factor to scale the logq correction, by default 1.0
    schema: Schema, optional
        The `Schema` with input features,
        by default None
    """

    def __init__(
        self,
        item_freq_probs: Union[tf.Tensor, Sequence],
        is_prob_distribution: bool = False,
        reg_factor: float = 1.0,
        schema: Schema = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if schema:
            self.set_schema(schema)

        self.reg_factor = reg_factor

        self._check_items_cardinality(item_freq_probs)
        candidate_probs = get_candidate_probs(item_freq_probs, is_prob_distribution)

        self.candidate_probs = tf.Variable(
            candidate_probs,
            name="candidate_probs",
            trainable=False,
            dtype=tf.float32,
            validate_shape=False,
            shape=tf.shape(candidate_probs),
        )

    @classmethod
    def from_parquet(
        cls,
        parquet_path: str,
        frequencies_probs_col: str,
        is_prob_distribution: bool = False,
        gpu: bool = True,
        schema: Schema = None,
        **kwargs,
    ):
        """Load the item frequency table from a parquet file
        (in the format automatically generated by NVTabular with workflow.fit()).
        It supposed the parquet file has a single column with the item frequencies
        and is indexed by item ids.

        Parameters
        ----------
        parquet_path : str
            Path to the parquet file
        frequencies_probs_col : str
            Column name containing the items frequencies / probabilities
        is_prob_distribution: bool, optional
            If True, the frequencies_probs_col should contain the probability
            distribution of the items. If False, the frequencies_probs_col values
            are frequencies and will be converted to probabilities
        gpu : bool, optional
            Whether to load data using cudf, by default True
        schema: Schema, optional
            The `Schema` with input features,
            by default None

        Returns
        -------
            An instance of PopularityLogitsCorrection
        """
        # TODO: Use the schema to infer the path to the item frequency parquet table
        if gpu:
            import cudf

            df = cudf.read_parquet(parquet_path)
            item_frequency = tf.squeeze(df_to_tensor(df[frequencies_probs_col]))
        else:
            import pandas as pd

            df = pd.read_parquet(parquet_path)
            item_frequency = tf.squeeze(tf.convert_to_tensor(df[frequencies_probs_col].values))
        return cls(
            item_freq_probs=item_frequency,
            is_prob_distribution=is_prob_distribution,
            schema=schema,
            **kwargs,
        )

    def get_candidate_probs(self):
        return self.candidate_probs.value()

    def update(
        self, item_freq_probs: Union[tf.Tensor, Sequence], is_prob_distribution: bool = False
    ):
        """Updates the item frequencies / probabilities

        Parameters:
        ----------
        item_freq_probs : Union[tf.Tensor, Sequence]
            A Tensor or list with item frequencies (if is_prob_distribution=False)
            or with item probabilities (if is_prob_distribution=True)
        is_prob_distribution: bool, optional
            If True, the item_freq_probs should be a probability distribution of the items.
            If False, the item frequencies is converted to probabilities
        """
        self._check_items_cardinality(item_freq_probs)
        candidate_probs = get_candidate_probs(item_freq_probs, is_prob_distribution)
        self.candidate_probs.assign(candidate_probs)

    def compute_output_shape(self, input_shape):
        return input_shape

    def call_outputs(
        self, outputs: PredictionOutput, training=True, **kwargs
    ) -> "PredictionOutput":
        predictions = outputs.predictions
        if training:
            positive_item_ids, negative_item_ids = (
                outputs.positive_item_ids,
                outputs.negative_item_ids,
            )
            positive_probs = tf.gather(self.candidate_probs, positive_item_ids)

            if negative_item_ids is not None:
                negative_probs = tf.gather(self.candidate_probs, negative_item_ids)
                # repeat negative scores for each positive item
                negative_probs = tf.reshape(
                    tf.tile(negative_probs, tf.shape(positive_item_ids)[0:1]),
                    (-1, tf.shape(negative_item_ids)[0]),
                )
                positive_probs = tf.concat(
                    [tf.expand_dims(positive_probs, -1), negative_probs], axis=1
                )

            # Applies the logQ correction
            epsilon = 1e-16
            predictions = predictions - (self.reg_factor * tf.math.log(positive_probs + epsilon))

        return outputs.copy_with_updates(predictions=predictions)

    def _check_items_cardinality(self, item_freq_probs):
        cardinalities = schema_utils.categorical_cardinalities(self.schema)
        item_id_feature_name = self.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
        if tf.shape(item_freq_probs)[0] != cardinalities[item_id_feature_name]:
            raise ValueError(
                "The item frequency table length does not match the item ids cardinality"
                f"(expected {cardinalities[item_id_feature_name]}"
                f", got {tf.shape(item_freq_probs)[0]})"
            )



# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List

import tensorflow as tf
from tensorflow.python.layers.base import Layer


class FIFOQueue(Layer):
    """Fixed size FIFO queue for caching input tensors
    over the batches. The storage is a fixed size tf.Variable()
    and some pointers are used to emulate the `queue()` / `dequeue()`
    ops. As soon as maximum capacity is reached, the oldest examples
    are removed from the queue to add the current beach ones.
    As an example use case, the `FIFOQueue` is used by
    `CachedCrossBatchSampler` to cache item embeddings and item metadata features.

        Parameters
        ----------
        capacity : int
            Maximum number of examples to store
        dtype : tf.DType
            The dtype of tensor that will be stored (e.g. tf.int32, tf.float32)
        dims : List[int], optional
            The dimension of the tensor examples, by default [], which means each example
            is a scalar.
        queue_name : str, optional
            This is string is concat with the tf.Variable() name to allow differentiating in the
            graph the FIFO Queue storages, by default ""
        initialize_tensor : tf.Tensor, optional
            Allows for initializing the storage with values, by default None which initializes
            the storage with -1.
            P.s. It is important that for categorical features the storage Variable is not
            initialized with a valid categorical value (e.g. values >= 0), so that `index_of()`
            works properly
    """

    def __init__(
        self,
        capacity: int,
        dtype: tf.DType,
        dims: List[int] = [],
        queue_name: str = "",
        initialize_tensor: tf.Tensor = None,
        **kwargs,
    ):
        assert capacity > 0

        super(FIFOQueue, self).__init__(**kwargs)
        self.capacity = capacity
        self.queue_dtype = dtype
        self.dims = dims
        self.queue_name = queue_name
        self.initialize_tensor = initialize_tensor

        self.first_pointer = tf.Variable(
            initial_value=tf.Variable(lambda: tf.zeros((), dtype=tf.int32)),
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            dtype=tf.int32,
        )
        self.next_available_pointer = tf.Variable(
            initial_value=tf.Variable(lambda: tf.zeros((), dtype=tf.int32)),
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            dtype=tf.int32,
        )
        self.at_full_capacity = tf.Variable(
            initial_value=tf.Variable(lambda: tf.zeros((), dtype=tf.bool)),
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            dtype=tf.bool,
        )

        if initialize_tensor is None:
            initialize_tensor = tf.Variable(
                lambda: tf.zeros([capacity] + self.dims, dtype=dtype) - 1
            )

        self.storage = tf.Variable(
            initial_value=initialize_tensor,
            name=f"{self.queue_name}/fifo_queue_storage",
            trainable=False,
            synchronization=tf.VariableSynchronization.NONE,
            shape=tf.TensorShape([self.capacity] + self.dims),
            dtype=self.queue_dtype,
        )

    def enqueue(self, val: tf.Tensor) -> None:
        """Enqueues an example into the queue

        Parameters
        ----------
        val : tf.Tensor
            Tensor to be stored in the queue. Its shape should match the `dims`
            set in the queue constructor
        """
        assert len(val.shape) == len(self.dims), "The rank of val and self.dims should match"
        assert list(val.shape) == self.dims, "The shape of val and self.dims should match"

        self.storage[self.next_available_pointer].assign(val)

        self.next_available_pointer.assign_add(1)
        if self.next_available_pointer >= self.capacity:
            self.next_available_pointer.assign(0)

        if self.at_full_capacity or self.next_available_pointer == self.first_pointer:
            self.first_pointer.assign(self.next_available_pointer)
            self.at_full_capacity.assign(True)

    def _check_input_values(self, values):
        if values.shape.dims is not None:
            assert len(tf.shape(values)) == len(self.dims) + 1, (
                "The rank of values (ignoring the first dim which is the number of examples) and "
                "self.dims should match"
            )
            assert list(values.shape[1:]) == self.dims, (
                "The shape of values (ignoring the first dim which is the number of examples) and "
                "self.dims should match"
            )

    def enqueue_many(self, vals: tf.Tensor) -> None:
        """Enqueues many examples into the queue.

        Parameters
        ----------
        val : tf.Tensor
            Tensor with the examples to be stored in the queue.
            The first dim of `vals` is the number of examples.
            From the second dim, its shape should match the `dims`
            set in the queue constructor
        """
        self._check_input_values(vals)

        # if values are larger than the queue capacity N, enqueueing only the last N items
        vals = vals[-self.capacity :]
        num_vals = int(tf.shape(vals)[0])

        next_pos_start = self.next_available_pointer
        next_pos_end = next_pos_start + num_vals
        if next_pos_end < self.capacity:
            self.storage[next_pos_start:next_pos_end].assign(vals)

            if self.at_full_capacity or (
                next_pos_start < self.first_pointer and next_pos_end >= self.first_pointer
            ):
                self.first_pointer.assign(next_pos_end)
                self.at_full_capacity.assign(True)
        else:
            num_overplus_items = next_pos_end - self.capacity
            next_pos_end = self.capacity
            self.storage[next_pos_start:next_pos_end].assign(vals[: num_vals - num_overplus_items])

            next_pos_start = 0
            next_pos_end = num_overplus_items
            self.storage[next_pos_start:next_pos_end].assign(vals[num_vals - num_overplus_items :])

            if self.at_full_capacity or next_pos_end >= self.first_pointer:
                self.first_pointer.assign(next_pos_end)
                self.at_full_capacity.assign(True)

        self.next_available_pointer.assign(next_pos_end)

    def dequeue(self) -> tf.Tensor:
        """Dequeues a single example from the queue

        Returns
        -------
        tf.Tensor
            A single example of the queue

        Raises
        ------
        IndexError
            The queue is empty
        """
        if self.first_pointer == self.next_available_pointer:
            raise IndexError("The queue is empty")
        self.at_full_capacity.assign(False)
        val = self.storage[self.first_pointer]
        self.first_pointer.assign_add(1)
        if self.first_pointer >= self.capacity:
            self.first_pointer.assign(0)
        return val

    def dequeue_many(self, n: int) -> tf.Tensor:
        """Dequeues many examples from the queue

        Parameters
        ----------
        n : int
            Number of examples to sample from the queue

        Returns
        -------
        tf.Tensor
            A tensor with N examples, being the first dim equal to N

        Raises
        ------
        IndexError
            The queue is empty
        ValueError
            The number of elements to dequeue must be greater than 0
        """
        if self.first_pointer == self.next_available_pointer:
            raise IndexError("The queue is empty")
        if n <= 0:
            raise ValueError("The number of elements to dequeue must be greater than 0.")
        self.at_full_capacity.assign(False)
        next_pos_start = self.first_pointer
        next_pos_end = next_pos_start + n

        if self.next_available_pointer > self.first_pointer:
            next_pos_end = min(next_pos_end, self.next_available_pointer)

        if next_pos_end < self.capacity:
            vals = self.storage[next_pos_start:next_pos_end]
        else:
            num_missing_items = next_pos_end - self.capacity
            next_pos_end = self.capacity
            vals1 = self.storage[next_pos_start:next_pos_end]

            next_pos_start = 0
            next_pos_end = min(num_missing_items, self.next_available_pointer)
            vals2 = self.storage[next_pos_start:next_pos_end]

            vals = tf.concat([vals1, vals2], axis=0)

        self.first_pointer.assign(next_pos_end)
        return vals

    def list_all(self) -> tf.Tensor:
        """Returns all items in the queue, sorted by the
        order they were added (FIFO)

        Returns
        -------
        tf.Tensor
            Returns a tensor with all examples added to the queue
        """
        if self.first_pointer < self.next_available_pointer:
            return self.storage[self.first_pointer : self.next_available_pointer]
        elif self.first_pointer == self.next_available_pointer and not self.at_full_capacity:
            return self.storage[0:0]  # Returns empty Tensor when queue is empty
        else:
            return tf.concat(
                [self.storage[self.first_pointer :], self.storage[: self.next_available_pointer]],
                axis=0,
            )

    def count(self) -> int:
        """Returns the number of examples added to the queue

        Returns
        -------
        int
            The number of examples added to the queue
        """
        if self.first_pointer < self.next_available_pointer:
            return self.next_available_pointer - self.first_pointer
        elif self.at_full_capacity:
            return self.capacity
        elif self.first_pointer == self.next_available_pointer:
            return 0
        else:
            return self.capacity - self.first_pointer + self.next_available_pointer

    def clear(self) -> None:
        """Removes all examples from the queue"""
        self.first_pointer.assign(0)
        self.next_available_pointer.assign(0)
        self.at_full_capacity.assign(False)

    def index_of(self, ids: tf.Tensor) -> tf.Tensor:
        """Retrieves the indices of the input ids if they exist in the queue,
        and -1 indicates the id was not found.

        Parameters
        ----------
        ids : tf.Tensor
            1D tensor with the search ids

        Returns
        -------
        tf.Tensor
            1D tensor with the same size of the input ids, containing the indices of the
            ids in the queue (-1 if not found)
        """
        assert self.queue_dtype in [tf.int8, tf.int16, tf.int32, tf.int64], (
            "The index_of method is only available for queues with an int dtype "
            "(tf.int8, tf.int16, tf.int32, tf.int64)"
        )
        assert (
            self.dims == []
        ), "The index_of method is only available for queues of scalars (dims=[])"

        # item_ids_indices = tf.where(tf.equal(ids, self.storage))
        equal_tensor = tf.cast(tf.equal(self.storage, tf.expand_dims(ids, -1)), tf.int32)
        # Artificially adding a first zero column so that index 0 is returned if
        # there is not an equal value (1) row-wise
        equal_tensor_extended = tf.concat(
            [tf.zeros(shape=(tf.shape(equal_tensor)[0], 1), dtype=tf.int32), equal_tensor],
            axis=1,
        )
        # Subtracting the indices to account for the first column added, so that
        # values not found become -1
        item_ids_indices = tf.argmax(equal_tensor_extended, axis=1) - 1
        return item_ids_indices

    def get_values_by_indices(self, indices: tf.Tensor) -> tf.Tensor:
        """Retrieves values of the queue based on their index

        Parameters
        ----------
        indices : tf.Tensor
            Indices to retrieve

        Returns
        -------
        tf.Tensor
            Values corresponding to the indices
        """
        result = tf.gather(self.storage, indices)
        return result

    def update_by_indices(self, indices: tf.Tensor, values: tf.Tensor) -> None:
        """Update values of the queue for specific indices

        Parameters
        ----------
        indices : tf.Tensor
            Indices to update
        values : tf.Tensor
            Update values
        """
        self._check_input_values(values)
        tf.assert_equal(
            tf.shape(indices)[0],
            tf.shape(values)[0],
            "The number of indices and values should match",
        )

        self.storage.scatter_nd_update(indices, values)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import math

import pytest
import tensorflow as tf
from sklearn.metrics import ndcg_score as ndcg_score_sklearn

from merlin.models.tf.metrics.ranking import (
    AvgPrecisionAt,
    MRRAt,
    NDCGAt,
    PrecisionAt,
    RecallAt,
    average_precision_at,
    dcg_at,
    mrr_at,
    ndcg_at,
    precision_at,
    recall_at,
)
from merlin.models.tf.utils.tf_utils import extract_topk


@pytest.fixture
def ranking_metrics_test_data():
    labels = tf.convert_to_tensor([[0, 1, 0, 1, 0], [1, 0, 0, 1, 0], [0, 0, 0, 0, 1]], tf.float32)
    predictions = tf.convert_to_tensor(
        [[10, 9, 8, 7, 6], [1, 4, 3, 2, 5], [10, 9, 8, 7, 6]], tf.float32
    )
    label_relevant_counts = tf.convert_to_tensor([2, 2, 0], tf.float32)
    return labels, predictions, label_relevant_counts


@pytest.fixture
def ranking_metrics_test_data_pre_sorted(ranking_metrics_test_data):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data
    predictions, labels, label_relevant_counts = extract_topk(5, predictions, labels)
    return labels, predictions, label_relevant_counts


def test_ranking_metrics_not_pre_sorted(ranking_metrics_test_data):
    labels, predictions, _ = ranking_metrics_test_data

    metric = RecallAt(k=4, pre_sorted=True)

    with pytest.raises(Exception) as excinfo:
        metric.update_state(labels, predictions)
    assert "you must provide label_relevant_counts argument" in str(excinfo.value)


def test_recall_at_k(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    result = recall_at(labels, predictions, label_relevant_counts, k=4)

    expected_result = [2 / 2, 1 / 2, 0]  # The last example has 0 relevant items
    tf.assert_equal(result, expected_result)


def test_precision_at_k(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    k = 4
    result = precision_at(labels, predictions, label_relevant_counts, k=k)

    expected_result = [2 / k, 1 / k, 0 / k]
    tf.assert_equal(result, expected_result)


def test_average_precision_at(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    result = average_precision_at(labels, predictions, label_relevant_counts, k=4)

    # Averaged precision at the position of relevant items among top-k,
    # divided by the total number of relevant items
    expected_result = [(1 / 2 + 2 / 4) / 2, (1 / 4) / 2, 0]  # The last example has 0 relevant items
    tf.assert_equal(result, expected_result)


def dcg_probe(pos_relevant, relevant_score=1):
    return relevant_score / math.log(pos_relevant + 1, 2)


def test_dcg_at(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    result = dcg_at(labels, predictions, label_relevant_counts, k=4)
    expected_result = [
        dcg_probe(2) + dcg_probe(4),
        dcg_probe(4),
        0,
    ]  # The last example has 0 relevant items
    tf.debugging.assert_near(result, expected_result)


def test_ndcg_at(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    result = ndcg_at(labels, predictions, label_relevant_counts, k=4)
    expected_result = [
        (dcg_probe(2) + dcg_probe(4)) / (dcg_probe(1) + dcg_probe(2)),
        dcg_probe(4) / (dcg_probe(1) + dcg_probe(2)),
        0,
    ]  # The last example has 0 relevant items
    tf.debugging.assert_near(result, expected_result)

    # Comparing with scikit learn
    result_sklearn = ndcg_score_sklearn(labels.numpy(), predictions.numpy(), k=4, ignore_ties=True)
    tf.debugging.assert_near(tf.cast(result_sklearn, tf.float32), tf.reduce_mean(expected_result))


def test_mrr_at(ranking_metrics_test_data_pre_sorted):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    result = mrr_at(labels, predictions, label_relevant_counts, k=4)
    expected_result = [1 / 2, 1 / 4, 0]  # The last example has 0 relevant items
    tf.debugging.assert_near(result, expected_result)


@pytest.mark.parametrize(
    "metric_exp_result",
    [
        (RecallAt(k=4), 0.5),
        (PrecisionAt(k=4), 0.25),
        (AvgPrecisionAt(k=4), 0.20833333),
        (MRRAt(k=4), 0.25),
        (NDCGAt(k=4), 0.30499637),
    ],
)
def test_ranking_metrics_classes(ranking_metrics_test_data_pre_sorted, metric_exp_result):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data_pre_sorted
    metric, expected_result = metric_exp_result

    metric.update_state(labels, predictions, label_relevant_counts)

    result = metric.result()
    tf.debugging.assert_near(result, expected_result)


@pytest.mark.parametrize(
    "metric_class",
    [
        RecallAt,
        PrecisionAt,
        AvgPrecisionAt,
        MRRAt,
        NDCGAt,
    ],
)
def test_ranking_metrics_classes_pre_or_not_sorted_matches(ranking_metrics_test_data, metric_class):
    labels, predictions, label_relevant_counts = ranking_metrics_test_data
    # Pre-sorting predictions and labels by predictions
    predictions_sorted, labels_sorted, label_relevant_counts_sorted = extract_topk(
        4, predictions, labels
    )

    metric1 = metric_class(k=4, pre_sorted=True)
    metric1.update_state(
        labels_sorted, predictions_sorted, label_relevant_counts=label_relevant_counts_sorted
    )
    result1 = metric1.result()

    metric2 = metric_class(k=4, pre_sorted=False)
    metric2.update_state(labels, predictions, label_relevant_counts=label_relevant_counts)
    result2 = metric2.result()

    tf.assert_equal(result1, result2)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Optional

import torch

from merlin.models.torch.features.base import InputBlock
from merlin.models.torch.tabular.base import (
    TABULAR_MODULE_PARAMS_DOCSTRING,
    FilterFeatures,
    TabularAggregationType,
    TabularTransformationType,
)
from merlin.models.utils.doc_utils import docstring_parameter
from merlin.schema import Schema


@docstring_parameter(tabular_module_parameters=TABULAR_MODULE_PARAMS_DOCSTRING)
class ContinuousFeatures(InputBlock):
    """Input block for continuous features.

    Parameters
    ----------
    features: List[str]
        List of continuous features to include in this module.
    {tabular_module_parameters}
    """

    def __init__(
        self,
        features: List[str],
        pre: Optional[TabularTransformationType] = None,
        post: Optional[TabularTransformationType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
    ):
        super().__init__(aggregation=aggregation, pre=pre, post=post, schema=schema)
        self.filter_features = FilterFeatures(features)

    @classmethod
    def from_features(cls, features, **kwargs):
        return cls(features, **kwargs)

    def forward(self, inputs, **kwargs):
        cont_features = self.filter_features(inputs)
        cont_features = {k: v.unsqueeze(-1) for k, v in cont_features.items()}
        return cont_features

    def forward_output_size(self, input_sizes):
        cont_features_sizes = self.filter_features.forward_output_size(input_sizes)
        cont_features_sizes = {k: torch.Size(list(v) + [1]) for k, v in cont_features_sizes.items()}
        return cont_features_sizes



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tensorflow as tf

from merlin.models.config.schema import FeatureCollection


class FeatureContext:
    def __init__(self, features: FeatureCollection, mask: tf.Tensor = None):
        self.features = features
        self._mask = mask

    @property
    def mask(self):
        if self._mask is None:
            raise ValueError("The mask is not stored, " "please make sure that a mask was set")
        return self._mask

    @mask.setter
    def mask(self, mask: tf.Tensor):
        self._mask = mask



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Union

import numpy as np
import tensorflow as tf
from tensorflow.python import to_dlpack

import merlin.io
from merlin.core.dispatch import DataFrameType
from merlin.models.tf.blocks.core.base import Block, PredictionOutput
from merlin.models.tf.utils import tf_utils
from merlin.models.tf.utils.batch_utils import TFModelEncode
from merlin.models.utils.constants import MIN_FLOAT
from merlin.schema import Tags


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class IndexBlock(Block):
    def __init__(self, values: tf.Tensor, ids: Optional[tf.Tensor] = None, **kwargs):
        super(IndexBlock, self).__init__(**kwargs)
        self.values = tf.Variable(
            values,
            name="values",
            trainable=False,
            dtype=tf.float32,
            validate_shape=False,
            shape=tf.TensorShape([None, tf.shape(values)[-1]]),
        )
        if ids is not None:
            id_dtype = ids.dtype
        else:
            id_dtype = tf.int64

        self.ids = tf.Variable(
            ids,
            name="ids",
            trainable=False,
            dtype=id_dtype,
            validate_shape=False,
            shape=tf.TensorShape([None]),
        )

    @classmethod
    def from_dataset(
        cls, data: merlin.io.Dataset, check_unique_ids: bool = True, **kwargs
    ) -> "IndexBlock":
        ids, values = cls.extract_ids_embeddings(data, check_unique_ids)
        return cls(values=values, ids=ids, **kwargs)

    @classmethod
    def from_block(
        cls, block: Block, data: merlin.io.Dataset, id_column: Optional[str] = None, **kwargs
    ) -> "IndexBlock":
        """Build candidates embeddings from applying `block` to a dataset of features `data`.

        Parameters:
        -----------
        block: Block
            The Block that returns embeddings from raw item features.
        data: merlin.io.Dataset
            Dataset containing raw item features.
        id_column: Optional[str]
            The candidates ids column name.
            Note, this will be inferred automatically if the block contains
            a schema with an item-id Tag.
        """
        embedding_df = cls.get_candidates_dataset(block, data, id_column)
        return cls.from_dataset(embedding_df, **kwargs)

    @staticmethod
    def _check_unique_ids(data: DataFrameType):
        if data.index.to_series().nunique() != data.shape[0]:
            raise ValueError("Please make sure that `data` contains unique indices")

    @classmethod
    def extract_ids_embeddings(cls, data: merlin.io.Dataset, check_unique_ids: bool = True):
        if hasattr(data, "to_ddf"):
            data = data.to_ddf()
        if check_unique_ids:
            cls._check_unique_ids(data=data)
        values = tf_utils.df_to_tensor(data)
        ids = tf_utils.df_to_tensor(data.index)

        if len(ids.shape) == 2:
            ids = tf.squeeze(ids)

        return ids, values

    @classmethod
    def get_candidates_dataset(
        cls, block: Block, data: merlin.io.Dataset, id_column: Optional[str] = None
    ):
        if not id_column and getattr(block, "schema", None):
            tagged = block.schema.select_by_tag(Tags.ITEM_ID)
            if tagged.column_schemas:
                id_column = tagged.first.name

        model_encode = TFModelEncode(model=block, output_concat_func=np.concatenate)

        data = data.to_ddf()
        embedding_ddf = data.map_partitions(model_encode, filter_input_columns=[id_column])
        embedding_df = embedding_ddf.compute(scheduler="synchronous")

        embedding_df.set_index(id_column, inplace=True)
        return embedding_df

    def update_from_block(
        self,
        block: Block,
        data: merlin.io.Dataset,
        id_column: Optional[str] = None,
        check_unique_ids: bool = True,
    ):
        embedding_df = IndexBlock.get_candidates_dataset(block, data, id_column)
        ids, embeddings = IndexBlock.extract_ids_embeddings(embedding_df, check_unique_ids)
        self.update(embeddings, ids)

    def update(self, values: tf.Tensor, ids: Optional[tf.Tensor] = None):
        if len(tf.shape(values)) != 2:
            raise ValueError(f"The candidates embeddings tensor must be 2D (got {values.shape}).")
        _ids: tf.Tensor = ids if ids is not None else tf.range(values.shape[0])
        self.ids.assign(_ids)
        self.values.assign(values)
        return self

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        return self.values[inputs]

    def to_dataset(self, gpu=True) -> merlin.io.Dataset:
        if gpu:
            import cudf

            df = cudf.from_dlpack(to_dlpack(tf.convert_to_tensor(self.values)))
            df.columns = [str(col) for col in list(df.columns)]
            df.set_index(cudf.RangeIndex(0, self.values.shape[0]))
        else:
            import pandas as pd

            df = pd.DataFrame(self.values.numpy())
            df.columns = [str(col) for col in list(df.columns)]
            df.set_index(pd.RangeIndex(0, self.values.shape[0]))

        return merlin.io.Dataset(df)


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class TopKIndexBlock(IndexBlock):
    """Top-K index to retrieve top-k scores and indices from an item block.

    Parameters:
    -----------
        k: int
            Number of top candidates to retrieve.
        values: tf.Tensor
            The pre-computed embedddings of candidates.
        ids: tf.Tensor
            The candidates ids.
    """

    def __init__(self, k, values: tf.Tensor, ids: Optional[tf.Tensor] = None, **kwargs):
        self._k = k
        super(TopKIndexBlock, self).__init__(values, ids, **kwargs)
        self.false_negatives_score = MIN_FLOAT

    @classmethod
    def from_block(  # type: ignore
        cls,
        block: Block,
        data: merlin.io.Dataset,
        k: int = 20,
        id_column: Optional[str] = None,
        **kwargs,
    ) -> "TopKIndexBlock":
        """
        class method to build candidates embeddings from
        applying `block` to a dataset of features `data`

        Parameters:
        -----------
        block: Block
            The Block that returns embeddings from raw item features.
        output_dim: int
            The output dimension of `block`.
        data: merlin.io.Dataset
            Dataset containing raw item features.
        k: int
            Number of top candidates to retrieve.
            Defaults to 20
        id_column: Optional[str]
            The candidates ids column name.
            Note, this will be inferred automatically if the block contains
            a schema with an item-id Tag.
        """
        return super().from_block(block=block, data=data, id_column=id_column, k=k, **kwargs)

    def call(self, inputs: tf.Tensor, k=None, **kwargs) -> Union[tf.Tensor, tf.Tensor]:
        """
        Compute Top-k scores and related indices from query inputs

        Parameters:
        ----------
        inputs: tf.Tensor
            Tensor of pre-computed query embeddings.
        k: int
            Number of top candidates to retrieve
            Defaults to constructor `_k` parameter.
        Returns
        -------
        top_scores, top_indices: tf.Tensor, tf.Tensor
            2D Tensors with the scores for the top-k candidates and related ids.
        """
        k = k if k is not None else self._k
        scores = tf.matmul(inputs, self.values, transpose_b=True)
        top_scores, top_indices = tf.math.top_k(scores, k=k)
        top_indices = tf.gather(self.ids, top_indices)

        return top_scores, top_indices

    def call_outputs(
        self, outputs: PredictionOutput, training=False, **kwargs
    ) -> "PredictionOutput":
        """
        Retrieve top-k negative scores for evaluation.

        Parameters
        ----------
        predictions: tf.Tensor
            Tensor of pre-computed positive scores.
            If`training=True`, the first column of predictions is expected
            to be positive scores and the remaining sampled negatives are ignored.

        Returns
        -------
        targets, predictions: tf.Tensor, tf.Tensor
            2D Tensors with the one-hot representation of true targets and
            the scores for the top-k implicit negatives.
        """
        queries = self.context["query"]
        pred_top_scores, top_ids = self(queries, k=self._k)

        targets_sorted = tf.cast(
            tf.expand_dims(outputs.positive_item_ids, -1) == top_ids, tf.float32
        )
        targets_sorted = tf.reshape(targets_sorted, tf.shape(pred_top_scores))

        label_relevant_counts = tf.ones([tf.shape(targets_sorted)[0]])

        return outputs.copy_with_updates(
            predictions=pred_top_scores,
            targets=targets_sorted,
            label_relevant_counts=label_relevant_counts,
        )

    def compute_output_shape(self, input_shape):
        batch_size = input_shape[0]
        return tf.TensorShape((batch_size, self._k)), tf.TensorShape((batch_size, self._k))



import os

import pytest

import merlin.io
from merlin.datasets.entertainment import get_movielens
from merlin.datasets.synthetic import generate_data

MAYBE_DATA_DIR = os.environ.get("INPUT_DATA_DIR", None)


def test_synthetic_music_data():
    dataset = generate_data("music-streaming", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 13


def test_movielens_25m_data():
    dataset = generate_data("movielens-25m", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 8


def test_movielens_1m_data():
    dataset = generate_data("movielens-1m", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 16


def test_movielens_100k_data():
    dataset = generate_data("movielens-100k", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 11


@pytest.mark.skipif(
    MAYBE_DATA_DIR is None,
    reason="No data-dir available, pass it through env variable $INPUT_DATA_DIR",
)
@pytest.mark.parametrize("variant", ["ml-25m", "ml-1m", "ml-100k"])
def test_get_movielens(tmp_path, variant):
    data_path = os.path.join(MAYBE_DATA_DIR, "movielens")

    train, valid = get_movielens(data_path, variant=variant, overwrite=True)

    assert isinstance(train, merlin.io.Dataset)
    assert isinstance(valid, merlin.io.Dataset)


def test_unsupported_variant():
    with pytest.raises(ValueError) as excinfo:
        get_movielens("/tmp", variant="unknown")
    assert "MovieLens dataset variant not supported" in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import warnings

try:
    from numba import cuda
except ImportError:
    cuda = None

try:
    import psutil
except ImportError:
    psutil = None


def _pynvml_mem_size(kind="total", index=0):
    import pynvml

    pynvml.nvmlInit()
    size = None
    if kind == "free":
        size = int(pynvml.nvmlDeviceGetMemoryInfo(pynvml.nvmlDeviceGetHandleByIndex(index)).free)
    elif kind == "total":
        size = int(pynvml.nvmlDeviceGetMemoryInfo(pynvml.nvmlDeviceGetHandleByIndex(index)).total)
    else:
        raise ValueError("{0} not a supported option for device_mem_size.".format(kind))
    pynvml.nvmlShutdown()
    return size


def device_mem_size(kind="total", cpu=False):

    # Use psutil (if available) for cpu mode
    if cpu and psutil:
        if kind == "total":
            return psutil.virtual_memory().total
        elif kind == "free":
            return psutil.virtual_memory().free
    elif cpu:
        warnings.warn("Please install psutil for full cpu=True support.")
        # Assume 1GB of memory
        return int(1e9)

    if kind not in ["free", "total"]:
        raise ValueError("{0} not a supported option for device_mem_size.".format(kind))
    try:
        if kind == "free":
            return int(cuda.current_context().get_memory_info()[0])
        else:
            return int(cuda.current_context().get_memory_info()[1])
    except NotImplementedError:
        if kind == "free":
            # Not using NVML "free" memory, because it will not include RMM-managed memory
            warnings.warn("get_memory_info is not supported. Using total device memory from NVML.")
        size = _pynvml_mem_size(kind="total", index=0)
    return size



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import partial

import numpy as np
import pytest
import torch

import merlin.models.torch as ml
from merlin.schema import Tags


def test_embedding_features(torch_cat_features):
    dim = 15
    feature_config = {
        f: ml.FeatureConfig(ml.TableConfig(100, dim, name=f)) for f in torch_cat_features.keys()
    }
    embeddings = ml.EmbeddingFeatures(feature_config)(torch_cat_features)

    assert list(embeddings.keys()) == list(feature_config.keys())
    assert all([emb.shape[-1] == dim for emb in embeddings.values()])


def test_embedding_features_layernorm(torch_cat_features):
    dim = 15
    feature_config = {
        f: ml.FeatureConfig(ml.TableConfig(100, dim, name=f)) for f in torch_cat_features.keys()
    }

    layer_norm = ml.TabularLayerNorm.from_feature_config(feature_config)
    embeddings = ml.EmbeddingFeatures(feature_config, post=layer_norm)(torch_cat_features)
    assert all(
        [emb.detach().numpy().mean() == pytest.approx(0.0, abs=0.1) for emb in embeddings.values()]
    )
    assert all([emb.detach().numpy().std() > 0.5 for emb in embeddings.values()])


def test_embedding_features_custom_init(torch_cat_features):
    MEAN = 1.0
    STD = 0.05
    emb_initializer = partial(torch.nn.init.normal_, mean=MEAN, std=STD)
    feature_config = {
        f: ml.FeatureConfig(ml.TableConfig(100, dim=15, name=f, initializer=emb_initializer))
        for f in torch_cat_features.keys()
    }
    embeddings = ml.EmbeddingFeatures(feature_config)(torch_cat_features)

    assert list(embeddings.keys()) == list(feature_config.keys())
    assert all(
        [emb.detach().numpy().mean() == pytest.approx(MEAN, abs=0.1) for emb in embeddings.values()]
    )
    assert all(
        [emb.detach().numpy().std() == pytest.approx(STD, abs=0.1) for emb in embeddings.values()]
    )


def test_table_config_invalid_embedding_initializer():
    with pytest.raises(ValueError) as excinfo:
        ml.TableConfig(100, dim=15, initializer="INVALID INITIALIZER")
    assert "initializer must be callable if specified" in str(excinfo.value)


def test_embedding_features_yoochoose(tabular_schema, torch_tabular_data):
    schema = tabular_schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(schema)
    embeddings = emb_module(torch_tabular_data)

    assert sorted(list(embeddings.keys())) == sorted(schema.column_names)
    assert all(emb.shape[-1] == 64 for emb in embeddings.values())
    assert emb_module.item_id == "item_id"

    max_value = list(schema.select_by_name("item_id"))[0].int_domain.max
    assert emb_module.item_embedding_table.num_embeddings == max_value + 1


def test_embedding_features_yoochoose_custom_dims(tabular_schema, torch_tabular_data):
    schema = tabular_schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(
        schema, embedding_dims={"item_id": 100}, embedding_dim_default=64
    )

    assert emb_module.embedding_tables["item_id"].weight.shape[1] == 100
    assert emb_module.embedding_tables["categories"].weight.shape[1] == 64

    embeddings = emb_module(torch_tabular_data)

    assert embeddings["item_id"].shape[1] == 100
    assert embeddings["categories"].shape[1] == 64


def test_embedding_features_yoochoose_infer_embedding_sizes(tabular_schema, torch_tabular_data):
    schema = tabular_schema.select_by_tag(Tags.CATEGORICAL)

    emb_module = ml.EmbeddingFeatures.from_schema(
        schema, infer_embedding_sizes=True, infer_embedding_sizes_multiplier=3.0
    )

    assert emb_module.embedding_tables["item_id"].weight.shape[1] == 46
    assert emb_module.embedding_tables["categories"].weight.shape[1] == 13

    embeddings = emb_module(torch_tabular_data)

    assert embeddings["item_id"].shape[1] == 46
    assert embeddings["categories"].shape[1] == 13


def test_embedding_features_yoochoose_custom_initializers(tabular_schema, torch_tabular_data):
    ITEM_MEAN = 1.0
    ITEM_STD = 0.05

    CATEGORY_MEAN = 2.0
    CATEGORY_STD = 0.1

    schema = tabular_schema.select_by_tag(Tags.CATEGORICAL)
    emb_module = ml.EmbeddingFeatures.from_schema(
        schema,
        layer_norm=False,
        embeddings_initializers={
            "item_id": partial(torch.nn.init.normal_, mean=ITEM_MEAN, std=ITEM_STD),
            "categories": partial(torch.nn.init.normal_, mean=CATEGORY_MEAN, std=CATEGORY_STD),
        },
    )

    embeddings = emb_module(torch_tabular_data)

    assert embeddings["item_id"].detach().numpy().mean() == pytest.approx(ITEM_MEAN, abs=0.1)
    assert embeddings["item_id"].detach().numpy().std() == pytest.approx(ITEM_STD, abs=0.1)

    assert embeddings["categories"].detach().numpy().mean() == pytest.approx(CATEGORY_MEAN, abs=0.1)
    assert embeddings["categories"].detach().numpy().std() == pytest.approx(CATEGORY_STD, abs=0.1)


def test_soft_embedding_invalid_num_embeddings():
    with pytest.raises(AssertionError) as excinfo:
        ml.SoftEmbedding(num_embeddings=0, embeddings_dim=16)
    assert "number of embeddings for soft embeddings needs to be greater than 0" in str(
        excinfo.value
    )


def test_soft_embedding_invalid_embeddings_dim():
    with pytest.raises(AssertionError) as excinfo:
        ml.SoftEmbedding(num_embeddings=10, embeddings_dim=0)
    assert "embeddings dim for soft embeddings needs to be greater than 0" in str(excinfo.value)


def test_soft_embedding():
    embeddings_dim = 16
    num_embeddings = 64

    soft_embedding = ml.SoftEmbedding(num_embeddings, embeddings_dim)
    assert soft_embedding.embedding_table.weight.shape == torch.Size(
        [num_embeddings, embeddings_dim]
    ), "Internal soft embedding table does not have the expected shape"

    batch_size = 10
    seq_length = 20
    cont_feature_inputs = torch.rand((batch_size, seq_length))
    output = soft_embedding(cont_feature_inputs)

    assert output.shape == torch.Size(
        [batch_size, seq_length, embeddings_dim]
    ), "Soft embedding output has not the expected shape"

    # Checking the default embedding initialization
    assert output.detach().numpy().mean() == pytest.approx(0.0, abs=0.1)
    assert output.detach().numpy().std() == pytest.approx(0.05, abs=0.2)


def test_soft_embedding_with_custom_init():
    embeddings_dim = 16
    num_embeddings = 64

    INIT_MEAN = 1.0
    INIT_STD = 0.05
    emb_initializer = partial(torch.nn.init.normal_, mean=INIT_MEAN, std=INIT_STD)
    soft_embedding = ml.SoftEmbedding(
        num_embeddings, embeddings_dim, emb_initializer=emb_initializer
    )
    assert soft_embedding.embedding_table.weight.shape == torch.Size(
        [num_embeddings, embeddings_dim]
    ), "Internal soft embedding table does not have the expected shape"

    batch_size = 10
    seq_length = 20
    cont_feature_inputs = torch.rand((batch_size, seq_length))
    output = soft_embedding(cont_feature_inputs)

    assert output.shape == torch.Size(
        [batch_size, seq_length, embeddings_dim]
    ), "Soft embedding output has not the expected shape"

    assert output.detach().numpy().mean() == pytest.approx(INIT_MEAN, abs=0.1)
    assert output.detach().numpy().std() == pytest.approx(INIT_STD, abs=0.1)


def test_soft_continuous_features(torch_con_features):
    dim = 16
    num_embeddings = 64

    emb_initializer = partial(torch.nn.init.normal_, mean=1.0, std=0.05)

    feature_config = {
        f: ml.FeatureConfig(
            ml.TableConfig(num_embeddings, dim, initializer=emb_initializer, name=f)
        )
        for f in torch_con_features.keys()
    }

    soft_embeddings = ml.SoftEmbeddingFeatures(feature_config)
    output = soft_embeddings(torch_con_features)

    assert list(output.keys()) == list(feature_config.keys())
    assert all(
        [list(v.shape) == list(torch_con_features[k].shape) + [dim] for k, v in output.items()]
    )


def test_layer_norm_features():
    ln = ml.TabularLayerNorm(features_dim={"a": 100, "b": 200})
    inputs = {
        "a": torch.tensor(np.random.uniform(1.0, 4.0, (500, 100)), dtype=torch.float32),
        "b": torch.tensor(np.random.uniform(2.0, 10.0, (500, 200)), dtype=torch.float32),
    }

    outputs = ln(inputs)

    assert all(
        [val.detach().numpy().mean() == pytest.approx(0.0, abs=0.1) for val in outputs.values()]
    )
    assert all(
        [val.detach().numpy().std() == pytest.approx(1.0, abs=0.1) for val in outputs.values()]
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
import inspect
from enum import Enum
from typing import Optional, Union

import tensorflow as tf
from tensorflow.python.keras import backend
from tensorflow.python.ops import nn

from merlin.models.config.schema import requires_schema
from merlin.models.tf.blocks.core.base import Block
from merlin.models.tf.blocks.core.tabular import TabularAggregation
from merlin.models.tf.typing import TabularData
from merlin.models.tf.utils import tf_utils
from merlin.models.utils.schema_utils import schema_to_tensorflow_metadata_json
from merlin.schema import Schema, Tags

# pylint has issues with TF array ops, so disable checks until fixed:
# https://github.com/PyCQA/pylint/issues/3613
# pylint: disable=no-value-for-parameter, unexpected-keyword-arg


@TabularAggregation.registry.register("concat")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ConcatFeatures(TabularAggregation):
    """Concatenates tensors along one dimension.

    Parameters
    ----------
    axis : int
        The axis to concatenate along.
    output_dtype : str
        The dtype of the output tensor.
    """

    def __init__(self, axis=-1, output_dtype=tf.float32, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.output_dtype = output_dtype

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        self._expand_non_sequential_features(inputs)
        self._check_concat_shapes(inputs)

        tensors = []
        for name in sorted(inputs.keys()):
            tensors.append(tf.cast(inputs[name], self.output_dtype))

        output = tf.concat(tensors, axis=-1)

        return output

    def compute_output_shape(self, input_shapes):
        agg_dim = sum([i[-1] for i in input_shapes.values()])
        if isinstance(agg_dim, tf.TensorShape):
            raise ValueError(f"Not possible to aggregate, received: {input_shapes}.")
        output_size = self._get_agg_output_size(input_shapes, agg_dim)
        return output_size

    def get_config(self):
        config = super().get_config()
        config["axis"] = self.axis
        config["output_dtype"] = self.output_dtype

        return config


@TabularAggregation.registry.register("stack")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class StackFeatures(TabularAggregation):
    """Stacks tensors along one dimension.

    Parameters
    ----------
    axis : int
        The axis to stack along.
    output_dtype : str
        The dtype of the output tensor.
    """

    def __init__(self, axis=-1, output_dtype=tf.float32, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.output_dtype = output_dtype

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        self._expand_non_sequential_features(inputs)
        self._check_concat_shapes(inputs)

        tensors = []
        for name in sorted(inputs.keys()):
            tensors.append(tf.cast(inputs[name], self.output_dtype))

        return tf.stack(tensors, axis=self.axis)

    def compute_output_shape(self, input_shapes):
        agg_dim = list(input_shapes.values())[0][-1]
        output_size = self._get_agg_output_size(input_shapes, agg_dim, axis=self.axis)

        if len(output_size) == 2:
            output_size = list(output_size)
            if self.axis == -1:
                output_size = [*output_size, len(input_shapes)]
            else:
                output_size.insert(self.axis, len(input_shapes))

        return output_size

    def get_config(self):
        config = super().get_config()
        config["axis"] = self.axis
        config["output_dtype"] = self.output_dtype

        return config


class ElementwiseFeatureAggregation(TabularAggregation):
    def _check_input_shapes_equal(self, inputs):
        all_input_shapes_equal = len(set([tuple(x.shape) for x in inputs.values()])) == 1
        if not all_input_shapes_equal:
            raise ValueError(
                "The shapes of all input features are not equal, which is required for element-wise"
                " aggregation: {}".format({k: v.shape for k, v in inputs.items()})
            )


@TabularAggregation.registry.register("sum")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class Sum(TabularAggregation):
    """Sum tensors along the first dimension."""

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        summed = tf.reduce_sum(list(inputs.values()), axis=0)

        return summed

    def compute_output_shape(self, input_shape):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_shape)
        last_dim = list(input_shape.values())[0][-1]

        return batch_size, last_dim


@TabularAggregation.registry.register("sum-residual")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SumResidual(Sum):
    def __init__(self, activation="relu", shortcut_name="shortcut", **kwargs):
        super().__init__(**kwargs)
        self.activation = tf.keras.activations.get(activation) if activation else None
        self.shortcut_name = shortcut_name

    def call(self, inputs: TabularData, **kwargs) -> Union[tf.Tensor, TabularData]:
        shortcut = inputs.pop(self.shortcut_name)
        outputs = {}
        for key, val in inputs.items():
            outputs[key] = tf.reduce_sum([inputs[key], shortcut], axis=0)
            if self.activation:
                outputs[key] = self.activation(outputs[key])

        if len(outputs) == 1:
            return list(outputs.values())[0]

        return outputs

    def compute_output_shape(self, input_shape):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_shape)
        last_dim = list(input_shape.values())[0][-1]

        return batch_size, last_dim

    def get_config(self):
        config = super().get_config()
        config["shortcut_name"] = self.shortcut_name
        config["activation"] = tf.keras.activations.serialize(self.activation)

        return config


@TabularAggregation.registry.register("add-left")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AddLeft(ElementwiseFeatureAggregation):
    """Concat tensors & add it to left_name"""

    def __init__(self, left_name="bias", **kwargs):
        super().__init__(**kwargs)
        self.left_name = left_name
        self.concat = ConcatFeatures()
        self.wide_logit = tf.keras.layers.Dense(1)

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        left = inputs.pop(self.left_name)
        if not left.shape[-1] == 1:
            left = self.wide_logit(left)

        return left + self.concat(inputs)

    def compute_output_shape(self, input_shape):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_shape)
        last_dim = list(input_shape.values())[0][-1]

        return batch_size, last_dim


@TabularAggregation.registry.register("element-wise-sum")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ElementwiseSum(ElementwiseFeatureAggregation):
    """Element-wise sum of all features."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.stack = StackFeatures(axis=0)

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        self._expand_non_sequential_features(inputs)
        self._check_input_shapes_equal(inputs)

        return tf.reduce_sum(self.stack(inputs), axis=0)

    def compute_output_shape(self, input_shape):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_shape)
        last_dim = list(input_shape.values())[0][-1]

        return batch_size, last_dim


@TabularAggregation.registry.register("element-wise-sum-item-multi")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
@requires_schema
class ElementwiseSumItemMulti(ElementwiseFeatureAggregation):
    """Element-wise sum of features."""

    def __init__(self, schema=None, **kwargs):
        super().__init__(**kwargs)
        self.stack = StackFeatures(axis=0)
        if schema:
            self.set_schema(schema)
        self.item_id_col_name = None

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        schema: Schema = self.schema  # type: ignore
        item_id_inputs = self.get_item_ids_from_inputs(inputs)
        self._expand_non_sequential_features(inputs)
        self._check_input_shapes_equal(inputs)

        item_id_column = schema.select_by_tag(Tags.ITEM_ID).first.name
        other_inputs = {k: v for k, v in inputs.items() if k != item_id_column}
        # Sum other inputs when there are multiple features.
        if len(other_inputs) > 1:
            other_inputs = tf.reduce_sum(self.stack(other_inputs), axis=0)
        else:
            other_inputs = list(other_inputs.values())[0]
        result = item_id_inputs * other_inputs
        return result

    def compute_output_shape(self, input_shape):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_shape)
        last_dim = list(input_shape.values())[0][-1]

        return batch_size, last_dim

    def get_config(self):
        config = super().get_config()
        if self.schema:
            config["schema"] = schema_to_tensorflow_metadata_json(self.schema)

        return config


class TupleAggregation(TabularAggregation, abc.ABC):
    """Aggregation between two features."""

    def call(self, left: tf.Tensor, right: tf.Tensor, **kwargs) -> tf.Tensor:  # type: ignore
        raise NotImplementedError()

    def super(self):
        return super()


# This is done this way to avoid mypy crashing.
def _tuple_aggregation_call(self, inputs: TabularData, *args, **kwargs):  # type: ignore
    if isinstance(inputs, tf.Tensor):
        left = inputs
        right = args[0]
    else:
        if not len(inputs) == 2:
            raise ValueError(f"Expected 2 inputs, got {len(inputs)}")
        left, right = tuple(inputs.values())
    outputs = self.super().__call__(left, right, **kwargs)  # type: ignore

    return outputs


TupleAggregation.__call__ = _tuple_aggregation_call  # type: ignore


@TabularAggregation.registry.register_with_multiple_names("cosine", "cosine-similarity")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class CosineSimilarity(TupleAggregation):
    """Cosine similarity aggregation"""

    def __init__(self, trainable=True, name=None, dtype=None, dynamic=False, **kwargs):
        super().__init__(trainable, name, dtype, dynamic, **kwargs)

    def call(self, left: tf.Tensor, right: tf.Tensor, **kwargs) -> tf.Tensor:  # type: ignore
        left = nn.l2_normalize(left, axis=1)
        right = nn.l2_normalize(right, axis=1)
        output = backend.batch_dot(left, right, 1)
        return output


@TabularAggregation.registry.register("elementwise-multiply")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ElementWiseMultiply(TupleAggregation):
    def __init__(self, trainable=True, name=None, dtype=None, dynamic=False, **kwargs):
        super().__init__(trainable, name, dtype, dynamic, **kwargs)

    def call(self, left: tf.Tensor, right: tf.Tensor, **kwargs) -> tf.Tensor:  # type: ignore
        out = tf.keras.layers.Multiply()([left, right])

        return out


def masked_mean(
    input_tensor: tf.Tensor, axis: Optional[int] = None, mask: tf.Tensor = None
) -> tf.Tensor:
    """Computes the mean of the specified axis considering only
    masked values

    Parameters
    ----------
    input_tensor : tf.Tensor
        Input tensor
    axis : int, optional
        axis to compute the mean over, by default None
    mask : tf.Tensor
        Mask tensor with the same shape as input_tensor.
        Only values where mask=True will be considered for average, by default None

    Returns
    -------
    tf.Tensor
        A tensor with the values averaged of the specified axis considering only masked values

    Raises
    ------
    ValueError
        If mask is not provided
    """
    output_tensor = input_tensor
    if mask is not None:
        mask_float = tf.expand_dims(tf.cast(mask, tf.float32), -1)
        output_tensor = tf.divide(
            tf.reduce_sum(tf.multiply(input_tensor, mask_float), axis=axis),
            tf.reduce_sum(mask_float, axis=axis),
        )
    else:
        raise ValueError("The mask is required for masked mean")

    return output_tensor


class SequenceAggregation(Enum):
    MEAN = tf.reduce_mean
    SUM = tf.reduce_sum
    MAX = tf.reduce_max
    MIN = tf.reduce_min
    MASKED_MEAN = masked_mean

    def __str__(self):
        return self.value

    def __eq__(self, o: object) -> bool:
        return str(o) == str(self)


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceAggregator(Block):
    """Computes the aggregation of elements across dimensions of a 3-D tensor.
    Args:
        combiner:
            tensorflow method to use for aggregation
            Defaults to SequenceAggregation.MEAN
        axis: int
            The dimensions to reduce.
            Defaults to 1
    """

    def __init__(self, combiner=SequenceAggregation.MEAN, axis: int = 1, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.combiner = combiner

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        assert len(inputs.shape) == 3, "inputs should be a 3-D tensor"
        kwargs = {}
        if (
            "mask" in inspect.signature(self.combiner).parameters
            and "valid_items_mask" in self.context.named_variables
        ):
            kwargs["mask"] = self.context["valid_items_mask"]

        return self.combiner(inputs, axis=self.axis, **kwargs)

    def compute_output_shape(self, input_shape):
        batch_size, _, last_dim = input_shape
        return batch_size, last_dim

    def get_config(self):
        config = super().get_config()
        config = tf_utils.maybe_serialize_keras_objects(
            self, config, {"combiner": tf.keras.layers.serialize}
        )
        return config

    @classmethod
    def from_config(cls, config):
        config = tf_utils.maybe_deserialize_keras_objects(
            config, ["combiner"], tf.keras.layers.deserialize
        )
        return super().from_config(config)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

import merlin.models.tf as ml
from merlin.io import Dataset
from merlin.models.config.schema import FeatureCollection
from merlin.models.tf import FeatureContext
from merlin.models.tf.blocks.core.base import PredictionOutput
from merlin.schema import Tags


@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_item_retrieval_scorer(ignore_last_batch_on_sample):
    batch_size = 10
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=ignore_last_batch_on_sample,
    )
    inbatch_sampler = ml.InBatchSampler()

    item_retrieval_scorer = ml.ItemRetrievalScorer(
        samplers=[cached_batches_sampler, inbatch_sampler],
        sampling_downscore_false_negatives=False,
        context=ml.ModelContext(),
    )

    users_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    items_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    positive_items = tf.random.uniform(shape=(10,), minval=1, maxval=100, dtype=tf.int32)

    # First batch
    output_scores1 = item_retrieval_scorer.call_outputs(
        PredictionOutput(
            {
                "query": users_embeddings,
                "item": items_embeddings,
            },
            targets=positive_items,
        ),
        training=True,
    ).predictions
    expected_num_samples_inbatch = batch_size
    expected_num_samples_cached = 0 if ignore_last_batch_on_sample else batch_size
    tf.assert_equal(tf.shape(output_scores1)[0], batch_size)
    # Number of negatives plus one positive
    tf.assert_equal(
        tf.shape(output_scores1)[1], expected_num_samples_inbatch + expected_num_samples_cached + 1
    )

    # Second batch
    output_scores2 = item_retrieval_scorer.call_outputs(
        PredictionOutput(
            {
                "query": users_embeddings,
                "item": items_embeddings,
            },
            targets=positive_items,
        ),
        training=True,
    ).predictions
    expected_num_samples_cached += batch_size
    tf.assert_equal(tf.shape(output_scores2)[0], batch_size)
    # Number of negatives plus one positive
    tf.assert_equal(
        tf.shape(output_scores2)[1],
        expected_num_samples_inbatch + expected_num_samples_cached + 1,
    )


""" @tf.function convert `call_outputs` of the ItemRetrievalScorer to graph ops,
In graph-model the exception is not raised.
or this test, we need to be able track the exceptions in graph mode.

def test_item_retrieval_scorer_cached_sampler_no_result_first_batch():
    batch_size = 10

    # CachedCrossBatchSampler is the only sampler here and with ignore_last_batch_on_sample=True
    # for the first batch no sample will be returned, which should raise an exception
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=True,
    )

    item_retrieval_scorer = ml.ItemRetrievalScorer(
        samplers=[cached_batches_sampler], sampling_downscore_false_negatives=False
    )

    users_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    items_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)

    with pytest.raises(Exception) as excinfo:
        _ = item_retrieval_scorer({"query": users_embeddings, "item": items_embeddings}, {})
    assert "No negative items where sampled from samplers" in str(excinfo.value)
"""


def test_item_retrieval_scorer_no_sampler():
    with pytest.raises(Exception) as excinfo:
        users_embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
        items_embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
        positive_items = tf.random.uniform(shape=(10,), minval=1, maxval=100, dtype=tf.int32)
        item_retrieval_scorer = ml.ItemRetrievalScorer(
            samplers=[], sampling_downscore_false_negatives=False, context=ml.ModelContext()
        )
        item_retrieval_scorer.call_outputs(
            PredictionOutput(
                {"query": users_embeddings, "item": items_embeddings}, targets=positive_items
            ),
            training=True,
        )
    assert "At least one sampler is required by ItemRetrievalScorer for negative sampling" in str(
        excinfo.value
    )


def test_item_retrieval_scorer_downscore_false_negatives():
    batch_size = 10

    cached_batches_sampler = ml.InBatchSampler()

    # Adding item id to the context
    item_ids = tf.random.uniform(shape=(batch_size,), minval=1, maxval=10000000, dtype=tf.int32)
    features = FeatureCollection(None, {"item_id": item_ids})
    feature_context = FeatureContext(features)

    FALSE_NEGATIVE_SCORE = -100_000_000.0
    item_retrieval_scorer = ml.ItemRetrievalScorer(
        samplers=[cached_batches_sampler],
        sampling_downscore_false_negatives=True,
        sampling_downscore_false_negatives_value=FALSE_NEGATIVE_SCORE,
    )

    users_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    items_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)

    outputs = item_retrieval_scorer.call_outputs(
        PredictionOutput({"query": users_embeddings, "item": items_embeddings}, {}),
        training=True,
        feature_context=feature_context,
    )
    output_scores = outputs.predictions

    output_neg_scores = output_scores[:, 1:]

    diag_mask = tf.eye(tf.shape(output_neg_scores)[0], dtype=tf.bool)
    tf.assert_equal(output_neg_scores[diag_mask], FALSE_NEGATIVE_SCORE)
    tf.assert_equal(
        tf.reduce_all(
            tf.not_equal(
                output_neg_scores[tf.math.logical_not(diag_mask)],
                tf.constant(FALSE_NEGATIVE_SCORE, dtype=output_neg_scores.dtype),
            )
        ),
        True,
    )


def test_item_retrieval_scorer_only_positive_when_not_training():
    batch_size = 10

    item_retrieval_scorer = ml.ItemRetrievalScorer(
        samplers=[ml.InBatchSampler()],
        sampling_downscore_false_negatives=False,
        context=ml.ModelContext(),
    )

    users_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    items_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)

    # Calls with training=False, so that only the positive item is scored
    output_scores = item_retrieval_scorer(
        {"query": users_embeddings, "item": items_embeddings}, training=False
    )
    tf.assert_equal(
        (int(tf.shape(output_scores)[0]), int(tf.shape(output_scores)[1])), (batch_size, 1)
    )


@pytest.mark.parametrize("run_eagerly", [True, False])
@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_retrieval_task_inbatch_cached_samplers(
    music_streaming_data: Dataset, run_eagerly, ignore_last_batch_on_sample
):
    batch_size = 100
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.TARGET)
    two_tower = ml.TwoTowerBlock(music_streaming_data.schema, query_tower=ml.MLPBlock([512, 256]))

    batch_inputs = ml.sample_batch(
        music_streaming_data, batch_size=batch_size, include_targets=False
    )

    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=ignore_last_batch_on_sample,
    )
    inbatch_sampler = ml.InBatchSampler()

    samplers = [inbatch_sampler, cached_batches_sampler]

    model = ml.Model(
        two_tower,
        ml.ItemRetrievalTask(
            music_streaming_data.schema,
            logits_temperature=2,
            samplers=samplers,
        ),
    )

    model.compile(optimizer="adam", run_eagerly=run_eagerly, loss="bpr")

    for batch_step in range(1, 4):
        output = model(batch_inputs, training=True)
        features = FeatureCollection(model.schema, model.as_dense(batch_inputs))
        feature_context = FeatureContext(features)
        output = (
            model.prediction_tasks[0]
            .pre.call_outputs(
                PredictionOutput(output, {}), training=True, feature_context=feature_context
            )
            .predictions
        )
        expected_num_samples_inbatch = batch_size
        expected_num_samples_cached = min(
            batch_size * (batch_step - 1 if ignore_last_batch_on_sample else batch_step),
            cached_batches_sampler.max_num_samples,
        )
        tf.assert_equal(tf.shape(output)[0], batch_size)
        # Number of negatives plus one positive
        tf.assert_equal(
            tf.shape(output)[1], expected_num_samples_inbatch + expected_num_samples_cached + 1
        )


@pytest.mark.parametrize("run_eagerly", [False])
def test_retrieval_task_inbatch_cached_samplers_fit(
    ecommerce_data: Dataset, run_eagerly, num_epochs=2
):
    ecommerce_data.schema = ecommerce_data.schema.remove_by_tag(Tags.TARGET)
    two_tower = ml.TwoTowerBlock(ecommerce_data.schema, query_tower=ml.MLPBlock([512, 256]))

    batch_size = 100

    samplers = [
        ml.InBatchSampler(),
        ml.CachedCrossBatchSampler(
            capacity=batch_size * 3,
            ignore_last_batch_on_sample=True,
        ),
        ml.CachedUniformSampler(
            capacity=batch_size * 3,
            ignore_last_batch_on_sample=False,
        ),
    ]
    task = ml.ItemRetrievalTask(
        ecommerce_data.schema,
        logits_temperature=2,
        samplers=samplers,
    )
    model = ml.RetrievalModel(two_tower, task)

    model.compile(optimizer="adam", run_eagerly=run_eagerly)
    losses = model.fit(ecommerce_data, batch_size=50, epochs=num_epochs)
    metrics = model.evaluate(
        x=ecommerce_data, batch_size=50, item_corpus=ecommerce_data, return_dict=True
    )

    assert len(losses.epoch) == num_epochs
    assert all(measure >= 0 for metric in losses.history for measure in losses.history[metric])
    assert set(metrics.keys()) == set(
        [
            "loss",
            "regularization_loss",
            "map_at_10",
            "mrr_at_10",
            "ndcg_at_10",
            "precision_at_10",
            "recall_at_10",
        ]
    )


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_retrieval_task_inbatch_cached_samplers_with_logits_correction(ecommerce_data, run_eagerly):
    from merlin.models.tf.blocks.core.transformations import PopularityLogitsCorrection

    batch_size = 100
    ecommerce_data.schema = ecommerce_data.schema.remove_by_tag(Tags.TARGET)

    two_tower = ml.TwoTowerBlock(ecommerce_data.schema, query_tower=ml.MLPBlock([512, 256]))
    samplers = [
        ml.InBatchSampler(),
        ml.CachedCrossBatchSampler(
            capacity=batch_size * 3,
            ignore_last_batch_on_sample=True,
        ),
        ml.CachedUniformSampler(
            capacity=batch_size * 3,
            ignore_last_batch_on_sample=False,
        ),
    ]
    NUM_ITEMS = 1001
    item_frequency = tf.sort(
        tf.random.uniform((NUM_ITEMS,), minval=0, maxval=NUM_ITEMS, dtype=tf.int32)
    )
    popularity_sampling_block = PopularityLogitsCorrection(
        item_frequency, schema=ecommerce_data.schema
    )

    task = ml.ItemRetrievalTask(
        ecommerce_data.schema,
        logits_temperature=2,
        samplers=samplers,
        post_logits=popularity_sampling_block,
        store_negative_ids=True,
    )
    model = ml.RetrievalModel(two_tower, task)

    # Setting up evaluation
    model.compile(optimizer="adam", run_eagerly=run_eagerly)

    losses = model.fit(ecommerce_data, batch_size=50, epochs=1)

    _ = model.evaluate(
        x=ecommerce_data, batch_size=50, item_corpus=ecommerce_data, return_dict=True
    )

    assert len(losses.epoch) == 1


@pytest.mark.parametrize("run_eagerly", [True])
@pytest.mark.parametrize("weight_tying", [True, False])
@pytest.mark.parametrize("sampled_softmax", [True, False])
def test_last_item_prediction_task(
    sequence_testing_data: Dataset,
    run_eagerly: bool,
    weight_tying: bool,
    sampled_softmax: bool,
):
    inputs = ml.InputBlock(
        sequence_testing_data.schema,
        aggregation="concat",
        seq=False,
        max_seq_length=4,
        masking="clm",
        split_sparse=True,
    )
    loss = tf.keras.losses.CategoricalCrossentropy(from_logits=True)
    task = ml.NextItemPredictionTask(
        schema=sequence_testing_data.schema,
        masking=True,
        weight_tying=weight_tying,
        sampled_softmax=sampled_softmax,
        logits_temperature=0.5,
    )

    model = ml.Model(inputs, ml.MLPBlock([64]), task)
    model.compile(optimizer="adam", run_eagerly=run_eagerly, loss=loss)
    losses = model.fit(sequence_testing_data, batch_size=50, epochs=2)

    assert len(losses.epoch) == 2
    for metric in losses.history.keys():
        assert type(losses.history[metric]) is list

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=50, include_targets=False, to_dense=True
    )
    out = model({k: tf.cast(v, tf.int64) for k, v in batch.items()})
    assert out.shape[-1] == 51997


@pytest.mark.parametrize("run_eagerly", [True, False])
@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_retrieval_task_inbatch_default_sampler(
    music_streaming_data: Dataset, run_eagerly, ignore_last_batch_on_sample
):
    batch_size = 100
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.TARGET)
    two_tower = ml.TwoTowerBlock(music_streaming_data.schema, query_tower=ml.MLPBlock([512, 256]))

    batch_inputs = ml.sample_batch(
        music_streaming_data, batch_size=batch_size, include_targets=False
    )

    model = ml.Model(
        two_tower,
        ml.ItemRetrievalTask(music_streaming_data.schema, logits_temperature=2),
    )

    model.compile(optimizer="adam", run_eagerly=run_eagerly, loss="bpr")

    for _ in range(1, 4):
        output = model(batch_inputs, training=True)
        features = FeatureCollection(model.schema, model.as_dense(batch_inputs))
        feature_context = FeatureContext(features)
        output = (
            model.prediction_tasks[0]
            .pre.call_outputs(
                PredictionOutput(output, {}), training=True, feature_context=feature_context
            )
            .predictions
        )
        expected_num_samples_inbatch = batch_size

        tf.assert_equal(tf.shape(output)[0], batch_size)
        # Number of negatives plus one positive
        tf.assert_equal(tf.shape(output)[1], expected_num_samples_inbatch + 1)



import pytest

import merlin.models.tf as ml
from merlin.io import Dataset


@pytest.mark.parametrize(
    "task_blocks",
    [
        None,
        ml.MLPBlock([32]),
        dict(click=ml.MLPBlock([16]), play_percentage=ml.MLPBlock([20])),
        dict(binary_classification_task=ml.MLPBlock([16]), regression_task=ml.MLPBlock([20])),
        {
            "click/binary_classification_task": ml.MLPBlock([16]),
            "play_percentage/regression_task": ml.MLPBlock([20]),
        },
    ],
)
def test_model_with_multiple_tasks(music_streaming_data: Dataset, task_blocks):
    music_streaming_data.schema = music_streaming_data.schema.without("like")

    inputs = ml.InputBlock(music_streaming_data.schema)
    prediction_tasks = ml.PredictionTasks(music_streaming_data.schema, task_blocks=task_blocks)
    model = ml.Model(inputs, ml.MLPBlock([64]), prediction_tasks)
    model.compile(optimizer="adam", run_eagerly=True)

    metrics = model.train_step(ml.sample_batch(music_streaming_data, batch_size=50))

    assert metrics["loss"] >= 0
    assert set(list(metrics.keys())) == set(
        [
            "loss",
            "regularization_loss",
            "click/binary_classification_task_loss",
            "play_percentage/regression_task_loss",
            "play_percentage/regression_task_root_mean_squared_error",
            "click/binary_classification_task_precision",
            "click/binary_classification_task_recall",
            "click/binary_classification_task_binary_accuracy",
            "click/binary_classification_task_auc",
        ]
    )
    if task_blocks:
        assert model.prediction_tasks[0].task_block != model.prediction_tasks[1].task_block


def test_mmoe_head(music_streaming_data: Dataset):
    inputs = ml.InputBlock(music_streaming_data.schema)
    prediction_tasks = ml.PredictionTasks(music_streaming_data.schema)
    mmoe = ml.MMOEBlock(prediction_tasks, expert_block=ml.MLPBlock([64]), num_experts=4)
    model = ml.Model(inputs, ml.MLPBlock([64]), mmoe, prediction_tasks)
    model.compile(optimizer="adam", run_eagerly=True)

    metrics = model.train_step(ml.sample_batch(music_streaming_data, batch_size=50))

    assert metrics["loss"] >= 0
    assert set(metrics.keys()) == set(
        [
            "loss",
            "click/binary_classification_task_loss",
            "like/binary_classification_task_loss",
            "play_percentage/regression_task_loss",
            "click/binary_classification_task_precision",
            "click/binary_classification_task_recall",
            "click/binary_classification_task_binary_accuracy",
            "click/binary_classification_task_auc",
            "like/binary_classification_task_precision_1",
            "like/binary_classification_task_recall_1",
            "like/binary_classification_task_binary_accuracy",
            "like/binary_classification_task_auc_1",
            "play_percentage/regression_task_root_mean_squared_error",
            "regularization_loss",
        ]
    )


def test_ple_head(music_streaming_data: Dataset):
    inputs = ml.InputBlock(music_streaming_data.schema)
    prediction_tasks = ml.PredictionTasks(music_streaming_data.schema)
    cgc = ml.CGCBlock(
        prediction_tasks, expert_block=ml.MLPBlock([64]), num_task_experts=2, num_shared_experts=2
    )
    model = ml.Model(inputs, ml.MLPBlock([64]), cgc, prediction_tasks)
    model.compile(optimizer="adam", run_eagerly=True)

    metrics = model.train_step(ml.sample_batch(music_streaming_data, batch_size=50))

    assert metrics["loss"] >= 0
    assert set(metrics.keys()) == set(
        [
            "loss",
            "click/binary_classification_task_loss",
            "like/binary_classification_task_loss",
            "play_percentage/regression_task_loss",
            "click/binary_classification_task_precision",
            "click/binary_classification_task_recall",
            "click/binary_classification_task_binary_accuracy",
            "click/binary_classification_task_auc",
            "like/binary_classification_task_precision_1",
            "like/binary_classification_task_recall_1",
            "like/binary_classification_task_binary_accuracy",
            "like/binary_classification_task_auc_1",
            "play_percentage/regression_task_root_mean_squared_error",
            "regularization_loss",
        ]
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

import merlin.models.tf as ml
from merlin.io import Dataset
from merlin.schema import Tags


def test_dlrm_block(testing_data: Dataset):
    dlrm = ml.DLRMBlock(
        testing_data.schema,
        embedding_dim=64,
        bottom_block=ml.MLPBlock([64]),
        top_block=ml.DenseResidualBlock(),
    )
    outputs = dlrm(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, 2080]


def test_dlrm_block_no_top_block(testing_data: Dataset):
    dlrm = ml.DLRMBlock(
        testing_data.schema,
        embedding_dim=64,
        bottom_block=ml.MLPBlock([64]),
    )
    outputs = dlrm(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, 2016]


def test_dlrm_block_no_continuous_features(testing_data: Dataset):
    schema = testing_data.schema.remove_by_tag(Tags.CONTINUOUS)
    dlrm = ml.DLRMBlock(schema, embedding_dim=64, top_block=ml.MLPBlock([32]))
    outputs = dlrm(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, 32]


def test_dlrm_block_no_categ_features(testing_data: Dataset):
    schema = testing_data.schema.remove_by_tag(Tags.CATEGORICAL)
    with pytest.raises(ValueError) as excinfo:
        ml.DLRMBlock(
            schema, embedding_dim=64, bottom_block=ml.MLPBlock([64]), top_block=ml.MLPBlock([16])
        )
    assert "DLRM requires categorical features" in str(excinfo.value)


def test_dlrm_block_single_categ_feature(testing_data: Dataset):
    schema = testing_data.schema.select_by_tag([Tags.ITEM_ID])
    dlrm = ml.DLRMBlock(schema, embedding_dim=64, top_block=ml.MLPBlock([32]))
    outputs = dlrm(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, 32]


def test_dlrm_block_no_schema():
    with pytest.raises(ValueError) as excinfo:
        ml.DLRMBlock(
            schema=None,
            embedding_dim=64,
            bottom_block=ml.MLPBlock([64]),
            top_block=ml.MLPBlock([32]),
        )
    assert "The schema is required by DLRM" in str(excinfo.value)


def test_dlrm_block_no_bottom_block(testing_data: Dataset):
    with pytest.raises(ValueError) as excinfo:
        ml.DLRMBlock(schema=testing_data.schema, embedding_dim=64, bottom_block=None)
    assert "The bottom_block is required by DLRM" in str(excinfo.value)


def test_dlrm_emb_dim_do_not_match_bottom_mlp(testing_data: Dataset):
    with pytest.raises(ValueError) as excinfo:
        ml.DLRMBlock(schema=testing_data.schema, bottom_block=ml.MLPBlock([64]), embedding_dim=75)
    assert "needs to match the last layer of bottom MLP" in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

import merlin.models.tf as ml
from merlin.io.dataset import Dataset
from merlin.models.tf.utils import testing_utils


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_simple_model(ecommerce_data: Dataset, run_eagerly):
    model = ml.Model(
        ml.InputBlock(ecommerce_data.schema),
        ml.MLPBlock([64]),
        ml.BinaryClassificationTask("click"),
    )

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_model_from_block(ecommerce_data: Dataset, run_eagerly):
    embedding_options = ml.EmbeddingOptions(embedding_dim_default=32)
    model = ml.Model.from_block(
        ml.MLPBlock([64]),
        ecommerce_data.schema,
        prediction_tasks=ml.BinaryClassificationTask("click"),
        embedding_options=embedding_options,
    )

    assert all(
        [f.table.dim == 32 for f in list(model.block.inputs["categorical"].feature_config.values())]
    )

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


def test_block_from_model_with_input(ecommerce_data: Dataset):
    inputs = ml.InputBlock(ecommerce_data.schema)
    block = inputs.connect(ml.MLPBlock([64]))

    with pytest.raises(ValueError) as excinfo:
        ml.Model.from_block(
            block,
            ecommerce_data.schema,
            input_block=inputs,
        )
    assert "The block already includes an InputBlock" in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# flake8: noqa

# Must happen before any importing of tensorflow to curtail mem usage
from merlin.models.loader.tf_utils import configure_tensorflow
from merlin.models.tf.utils.tf_utils import TensorInitializer

configure_tensorflow()

from tensorflow.keras.layers import Dense, Layer
from tensorflow.python.keras.losses import Loss
from tensorflow.python.keras.metrics import Metric
from tensorflow.python.keras.optimizer_v2.optimizer_v2 import OptimizerV2
from tensorflow.python.training.tracking.data_structures import ListWrapper, _DictWrapper

from merlin.models.loader.tf_utils import configure_tensorflow
from merlin.models.tf.blocks.core.aggregation import (
    ConcatFeatures,
    ElementwiseSum,
    ElementwiseSumItemMulti,
    StackFeatures,
)
from merlin.models.tf.blocks.core.base import (
    Block,
    EmbeddingWithMetadata,
    ModelContext,
    NoOp,
    right_shift_layer,
)
from merlin.models.tf.blocks.core.combinators import ParallelBlock, ResidualBlock, SequentialBlock
from merlin.models.tf.blocks.core.context import FeatureContext
from merlin.models.tf.blocks.core.index import IndexBlock, TopKIndexBlock
from merlin.models.tf.blocks.core.masking import CausalLanguageModeling, MaskedLanguageModeling
from merlin.models.tf.blocks.core.tabular import AsTabular, Filter, TabularBlock
from merlin.models.tf.blocks.core.transformations import (
    AsDenseFeatures,
    AsSparseFeatures,
    CategoricalOneHot,
    ExpandDims,
    LabelToOneHot,
    StochasticSwapNoise,
)
from merlin.models.tf.blocks.cross import CrossBlock
from merlin.models.tf.blocks.dlrm import DLRMBlock
from merlin.models.tf.blocks.experts import CGCBlock, MMOEBlock, MMOEGate
from merlin.models.tf.blocks.interaction import DotProductInteraction, FMPairwiseInteraction
from merlin.models.tf.blocks.mlp import DenseResidualBlock, MLPBlock
from merlin.models.tf.blocks.retrieval.base import DualEncoderBlock, ItemRetrievalScorer
from merlin.models.tf.blocks.retrieval.matrix_factorization import (
    MatrixFactorizationBlock,
    QueryItemIdsEmbeddingsBlock,
)
from merlin.models.tf.blocks.retrieval.two_tower import TwoTowerBlock
from merlin.models.tf.blocks.sampling.base import ItemSampler
from merlin.models.tf.blocks.sampling.cross_batch import (
    CachedCrossBatchSampler,
    CachedUniformSampler,
    PopularityBasedSampler,
)
from merlin.models.tf.blocks.sampling.in_batch import InBatchSampler
from merlin.models.tf.blocks.sampling.queue import FIFOQueue
from merlin.models.tf.dataset import sample_batch
from merlin.models.tf.inputs.base import InputBlock
from merlin.models.tf.inputs.continuous import ContinuousFeatures
from merlin.models.tf.inputs.embedding import (
    ContinuousEmbedding,
    EmbeddingFeatures,
    EmbeddingOptions,
    FeatureConfig,
    SequenceEmbeddingFeatures,
    TableConfig,
)
from merlin.models.tf.losses import LossType
from merlin.models.tf.metrics.ranking import (
    AvgPrecisionAt,
    MRRAt,
    NDCGAt,
    PrecisionAt,
    RecallAt,
    ranking_metrics,
)
from merlin.models.tf.models import benchmark
from merlin.models.tf.models.base import Model, RetrievalModel
from merlin.models.tf.models.ranking import DCNModel, DeepFMModel, DLRMModel
from merlin.models.tf.models.retrieval import (
    MatrixFactorizationModel,
    TwoTowerModel,
    YoutubeDNNRetrievalModel,
)
from merlin.models.tf.prediction_tasks.base import ParallelPredictionBlock, PredictionTask
from merlin.models.tf.prediction_tasks.classification import (
    BinaryClassificationTask,
    MultiClassClassificationTask,
)
from merlin.models.tf.prediction_tasks.multi import PredictionTasks
from merlin.models.tf.prediction_tasks.next_item import NextItemPredictionTask
from merlin.models.tf.prediction_tasks.regression import RegressionTask
from merlin.models.tf.prediction_tasks.retrieval import ItemRetrievalTask
from merlin.models.tf.utils import repr_utils

ListWrapper.__repr__ = repr_utils.list_wrapper_repr
_DictWrapper.__repr__ = repr_utils.dict_wrapper_repr

Dense.repr_extra = repr_utils.dense_extra_repr
Layer.__rrshift__ = right_shift_layer
Layer.__repr__ = repr_utils.layer_repr
Loss.__repr__ = repr_utils.layer_repr_no_children
Metric.__repr__ = repr_utils.layer_repr_no_children
OptimizerV2.__repr__ = repr_utils.layer_repr_no_children

__all__ = [
    "Block",
    "ModelContext",
    "SequentialBlock",
    "ResidualBlock",
    "DualEncoderBlock",
    "CrossBlock",
    "DLRMBlock",
    "MLPBlock",
    "CausalLanguageModeling",
    "MaskedLanguageModeling",
    "ContinuousEmbedding",
    "MMOEGate",
    "MMOEBlock",
    "CGCBlock",
    "TopKIndexBlock",
    "IndexBlock",
    "FeatureContext",
    "DenseResidualBlock",
    "TabularBlock",
    "ContinuousFeatures",
    "EmbeddingFeatures",
    "SequenceEmbeddingFeatures",
    "EmbeddingOptions",
    "FeatureConfig",
    "TableConfig",
    "ParallelPredictionBlock",
    "TwoTowerBlock",
    "MatrixFactorizationBlock",
    "QueryItemIdsEmbeddingsBlock",
    "AsDenseFeatures",
    "AsSparseFeatures",
    "CategoricalOneHot",
    "ElementwiseSum",
    "ElementwiseSumItemMulti",
    "AsTabular",
    "ConcatFeatures",
    "Filter",
    "ParallelBlock",
    "StackFeatures",
    "DotProductInteraction",
    "FMPairwiseInteraction",
    "LabelToOneHot",
    "PredictionTask",
    "BinaryClassificationTask",
    "MultiClassClassificationTask",
    "RegressionTask",
    "ItemRetrievalTask",
    "ItemRetrievalScorer",
    "NextItemPredictionTask",
    "NDCGAt",
    "PrecisionAt",
    "MRRAt",
    "AvgPrecisionAt",
    "RecallAt",
    "ranking_metrics",
    "Model",
    "RetrievalModel",
    "InputBlock",
    "PredictionTasks",
    "StochasticSwapNoise",
    "ExpandDims",
    "NoOp",
    "ItemSampler",
    "EmbeddingWithMetadata",
    "InBatchSampler",
    "CachedCrossBatchSampler",
    "CachedUniformSampler",
    "PopularityBasedSampler",
    "FIFOQueue",
    "YoutubeDNNRetrievalModel",
    "TwoTowerModel",
    "MatrixFactorizationModel",
    "DLRMModel",
    "DCNModel",
    "DeepFMModel",
    "losses",
    "LossType",
    "sample_batch",
    "TensorInitializer",
]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from typing import Optional

import tensorflow as tf

from merlin.models.tf.blocks.core.base import Block, BlockType
from merlin.models.tf.blocks.retrieval.base import DualEncoderBlock, RetrievalMixin
from merlin.models.tf.inputs.base import InputBlock
from merlin.models.tf.inputs.embedding import EmbeddingOptions
from merlin.schema import Schema, Tags

LOG = logging.getLogger("merlin_models")


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class TwoTowerBlock(DualEncoderBlock, RetrievalMixin):
    """
    Builds the Two-tower architecture, as proposed in the following
    `paper https://doi.org/10.1145/3298689.3346996`_ [Xinyang19].

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    query_tower : Block
        The `Block` that combines user features
    item_tower : Optional[Block], optional
        The optional `Block` that combines items features.
        If not provided, a copy of the query_tower is used.
    query_tower_tag : Tag
        The tag to select query features, by default `Tags.USER`
    item_tower_tag : Tag
        The tag to select item features, by default `Tags.ITEM`
    embedding_options : EmbeddingOptions
        Options for the input embeddings.
        - embedding_dims: Optional[Dict[str, int]] - The dimension of the
        embedding table for each feature (key), by default None
        - embedding_dim_default: int - Default dimension of the embedding
        table, when the feature is not found in ``embedding_dims``, by default 64
        - infer_embedding_sizes : bool, Automatically defines the embedding
        dimension from the feature cardinality in the schema, by default False
        - infer_embedding_sizes_multiplier: int. Multiplier used by the heuristic
        to infer the embedding dimension from its cardinality. Generally
        reasonable values range between 2.0 and 10.0. By default 2.0.
    post: Optional[Block], optional
        The optional `Block` to apply on both outputs of Two-tower model

    Returns
    -------
    ParallelBlock
        The Two-tower block

    Raises
    ------
    ValueError
        The schema is required by TwoTower
    ValueError
        The query_tower is required by TwoTower
    """

    def __init__(
        self,
        schema: Schema,
        query_tower: Block,
        item_tower: Optional[Block] = None,
        query_tower_tag=Tags.USER,
        item_tower_tag=Tags.ITEM,
        embedding_options: EmbeddingOptions = EmbeddingOptions(
            embedding_dims=None,
            embedding_dim_default=64,
            infer_embedding_sizes=False,
            infer_embedding_sizes_multiplier=2.0,
        ),
        post: Optional[BlockType] = None,
        **kwargs,
    ):
        if schema is None:
            raise ValueError("The schema is required by TwoTower")
        if query_tower is None:
            raise ValueError("The query_tower is required by TwoTower")

        _item_tower: Block = item_tower or query_tower.copy()
        if not getattr(_item_tower, "inputs", None):
            item_schema = schema.select_by_tag(item_tower_tag) if item_tower_tag else schema
            if not item_schema:
                raise ValueError(
                    f"The schema should contain features with the tag `{item_tower_tag}`,"
                    "required by item-tower"
                )
            item_tower_inputs = InputBlock(item_schema, embedding_options=embedding_options)
            _item_tower = item_tower_inputs.connect(_item_tower)
        if not getattr(query_tower, "inputs", None):
            query_schema = schema.select_by_tag(query_tower_tag) if query_tower_tag else schema
            if not query_schema:
                raise ValueError(
                    f"The schema should contain features with the tag `{query_schema}`,"
                    "required by query-tower"
                )
            query_inputs = InputBlock(query_schema, embedding_options=embedding_options)
            query_tower = query_inputs.connect(query_tower)

        super().__init__(query_tower, _item_tower, post=post, **kwargs)



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import implicit
from implicit.evaluation import ranking_metrics_at_k

from merlin.io import Dataset
from merlin.models.utils.dataset import _to_numpy, dataset_to_coo
from merlin.schema import Tags


class ImplicitModelAdaptor:
    """
    This class adapts a model from implicit to work with the high level merlin-models api

    Example usage::

        # Get the movielens dataset
        from merlin.datasets.entertainment import get_movielens

        train, valid = get_movielens()

        # Train an ALS model with implicit using the merlin movielens dataset
        from merlin.models.implicit import AlternatingLeastSquares

        model = AlternatingLeastSquares(factors=128, iterations=15, regularization=0.01)
        model.fit(train)

        # evaluate the model given the validation set
        # prints out {'precision@10': 0.3895182175113377, 'map@10': 0.2609445966914911, ...} etc
        print(model.evaluate(valid))
    """

    def __init__(self, implicit_model):
        self.implicit_model = implicit_model
        self.train_data = None

    def fit(self, train: Dataset):
        """Trains the implicit model

        Parameters
        ----------
        train : merlin.io.Dataset
            The training dataset to use to fit the model. We will use the column tagged
            merlin.schema.Tags.ITEM_ID as the item , and merlin.schema.Tags.USER_ID as the userid.
            If there is a column tagged as Tags.TARGET we will also use that for the values,
            otherwise will be set to 1
        """
        data = dataset_to_coo(train).tocsr()
        self.implicit_model.fit(data)
        self.train_data = data

    def evaluate(self, test_dataset: Dataset, k=10):
        """Evaluates the model

        This function evaluates using a variety of ranking metrics, and returns
        a dictionary of {metric_name: value}.

        Parameters
        ----------
        test_dataset : merlin.io.Dataset
            The validation dataset to evaluate
        k : int
            How many items to return per prediction. By default this method will
            return metrics like 'map@10' , but by increasing k you can generate
            different versions
        """
        test = dataset_to_coo(test_dataset).tocsr()
        ret = ranking_metrics_at_k(
            self.implicit_model,
            self.train_data,
            test,
            K=k,
        )
        return {metric + f"@{k}": value for metric, value in ret.items()}

    def predict(self, dataset: Dataset, k=10):
        """Generate predictions from the dataset

        Parameters
        ----------
        test_dataset : merlin.io.Dataset
            The dataset to use for generating predictions. Each userid (as denoted by Tags.USER_ID
            will get recommendations generated for them
        k: int
            The number of recommendations to generate for each user
        """
        # Get the user-ids for the dataset,
        user_id_column = dataset.schema.select_by_tag(Tags.USER_ID).first.name
        userids = _to_numpy(
            dataset.to_ddf()[user_id_column].unique().compute(scheduler="synchronous")
        )

        return self.implicit_model.recommend(userids, None, filter_already_liked_items=False, N=k)


class AlternatingLeastSquares(ImplicitModelAdaptor):
    def __init__(self, *args, **kwargs):
        super().__init__(implicit.als.AlternatingLeastSquares(*args, **kwargs))


class BayesianPersonalizedRanking(ImplicitModelAdaptor):
    def __init__(self, *args, **kwargs):
        super().__init__(implicit.bpr.BayesianPersonalizedRanking(*args, **kwargs))



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pathlib

import numpy as np
import pytest
import tensorflow as tf

NUM_EXAMPLES = 1000
MAX_CARDINALITY = 100
ASSETS_DIR = pathlib.Path(__file__).parent.parent / "assets"


@pytest.fixture(autouse=True)
def tf_clear_session():
    tf.keras.backend.clear_session()


@pytest.fixture
def tf_con_features():
    features = {}
    keys = "abcdef"

    for key in keys:
        features[key] = tf.random.uniform((NUM_EXAMPLES, 1))

    return features


@pytest.fixture
def tf_cat_features():
    features = {}
    keys = [f"cat_{f}" for f in "abcdef"]

    for key in keys:
        features[key] = tf.random.uniform((NUM_EXAMPLES, 1), maxval=MAX_CARDINALITY, dtype=tf.int32)

    return features


@pytest.fixture
def tf_masking_inputs():
    # fixed parameters for tests
    NUM_EXAMPLES = 20
    MAX_LEN = 10
    PAD_TOKEN = 0
    hidden_dim = 16
    features = {}
    # generate random tensors for test
    features["input_tensor"] = tf.convert_to_tensor(
        np.random.uniform(0, 1, (NUM_EXAMPLES, MAX_LEN, hidden_dim))
    )
    # create sequences
    labels = np.random.randint(1, MAX_CARDINALITY, (NUM_EXAMPLES, MAX_LEN))
    # replace last 2 items by zeros to mimic padding
    labels[:, MAX_LEN - 2 :] = 0
    labels = tf.convert_to_tensor(labels)
    features["labels"] = labels
    features["padding_idx"] = PAD_TOKEN
    features["vocab_size"] = MAX_CARDINALITY

    return features


@pytest.fixture
def tf_ranking_metrics_inputs():
    POS_EXAMPLE = 30
    VOCAB_SIZE = 40
    features = {}
    features["scores"] = tf.convert_to_tensor(np.random.uniform(0, 1, (POS_EXAMPLE, VOCAB_SIZE)))
    features["ks"] = tf.convert_to_tensor([1, 2, 3, 5, 10, 20])
    features["labels_one_hot"] = tf.convert_to_tensor(
        np.random.choice(a=[0, 1], size=(POS_EXAMPLE, VOCAB_SIZE))
    )

    features["labels"] = tf.convert_to_tensor(np.random.randint(1, VOCAB_SIZE, (POS_EXAMPLE,)))
    return features



import warnings
from typing import Dict, List, Optional, Union

import distributed
import numpy as np
import xgboost as xgb

from merlin.core.utils import global_dask_client
from merlin.io import Dataset
from merlin.schema import Schema, Tags


class XGBoost:
    """Create an XGBoost model from a merlin dataset.
    The class adapts an XGBoost model to work with the high level merlin-models API.

    Example usage::

        # get the movielens dataset
        from merlin.datasets.entertainment import get_movielens

        train, valid = get_movielens(variant="ml-1m")

        # Train an XGBoost model
        from merlin.core.utils import Distributed
        from merlin.models.xgb import XGBoost

        with Distributed():
            model = XGBoost(train.schema, objective="binary:logistic")
            model.fit(train)
            metrics = model.evaluate(valid)
    """

    def __init__(
        self,
        schema: Schema,
        *,
        target_columns: Optional[Union[str, list]] = None,
        qid_column: Optional[str] = None,
        objective: str = "reg:squarederror",
        booster: Optional[xgb.Booster] = None,
        **params,
    ):
        """
        Parameters
        ----------
        schema : merlin.schema.Schema
            The schema of the data that will be used to train and evaluate the model.
        target_columns : Optional[Union[list, str]]
            The target columns to use. If provided, will be used as the label(s).
            Otherwise the targets are automatically inferred from the objective and column tags.
        qid_column : Optional[str]
            For ranking objectives. The query ID column. If not provided will use
            the user ID (tagged with merlin.schema.Tags.USER_ID) column.
        objective : str
            The XGBoost objective to use. List of XGBoost objective functions:
            https://xgboost.readthedocs.io/en/stable/gpu/index.html#objective-functions
        **params
            The parameters to use for the XGBoost train method
        """
        self.params = {**params, "objective": objective}

        target_tag = get_target_tag(objective)
        self.target_columns = target_columns or get_targets(schema, target_tag)

        if objective.startswith("rank") and qid_column is None:
            qid_column = schema.select_by_tag(Tags.USER_ID).column_names[0]
        self.qid_column = qid_column

        self.booster = booster

    @property
    def dask_client(self) -> Optional[distributed.Client]:
        return global_dask_client()

    def fit(
        self,
        train: Dataset,
        **train_kwargs,
    ) -> xgb.Booster:
        """Trains the XGBoost Model.

        Will use the columns tagged with the target_type passed to the
        constructor as the labels.  And all other non-list columns as
        input features.

        Parameters
        ----------
        train : merlin.io.Dataset
            The training dataset to use to fit the model.
            We will use the column(s) tagged with merlin.schema.Tags.TARGET that match the
            objective as the label(s).
        **train_kwargs
            Additional keyword arguments passed to the xgboost.train function

        Returns
        -------
        xgb.Booster

        Raises
        ------
        ValueError
           If objective is not supported. Or if the target columns cannot be found.
        """
        X, y, qid = dataset_to_xy(
            train,
            self.target_columns,
            self.qid_column,
        )

        dtrain = xgb.dask.DaskDMatrix(self.dask_client, X, label=y, qid=qid)
        watchlist = [(dtrain, "train")]

        booster: xgb.Booster = xgb.dask.train(
            self.dask_client, self.params, dtrain, evals=watchlist, **train_kwargs
        )["booster"]
        self.booster = booster

        return booster

    def evaluate(self, dataset: Dataset, **predict_kwargs) -> Dict[str, float]:
        """Evaluates the model on the dataset provided.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The dataset used to evaluate the model.
        **predict_kwargs
            The keyword parameters passed to the predict function

        Returns
        -------
        Dict[str, float]
            Dictionary of metrics of the form {metric_name: value}.
        """
        if self.booster is None:
            raise ValueError("The fit method must be called before evaluate.")

        X, _, qid = dataset_to_xy(dataset, self.target_columns, self.qid_column)
        data = xgb.dask.DaskDMatrix(self.dask_client, X, qid=qid)
        preds = xgb.dask.predict(self.dask_client, self.booster, data, **predict_kwargs)

        # convert to DMatrix
        # (eval doesn't have dask support currently)
        if qid is not None:
            qid = qid.compute()
        eval_data = xgb.DMatrix(X.compute(), label=preds.compute(), qid=qid)

        metrics_str = self.booster.eval(eval_data)
        metrics = {}
        for metric in metrics_str.split("\t")[1:]:
            metric_name, metric_value = metric.split(":")
            metrics[metric_name[len("eval-") :]] = float(metric_value)

        return metrics

    def predict(self, dataset: Dataset, **predict_kwargs) -> np.ndarray:
        """Generate predictions from the dataset.

        Parameters
        ----------
        dataset : merlin.io.Dataset
            The dataset to use for predictions
        **predict_kwargs
            keyword arguments passed to the xgboost.core.Booster.predict method

        Returns
        -------
        numpy.ndarray
            The predicions data
        """
        if self.booster is None:
            raise ValueError("The fit method must be called before predict.")

        X, _, qid = dataset_to_xy(dataset, self.target_columns, self.qid_column)
        data = xgb.dask.DaskDMatrix(self.dask_client, X, qid=qid)
        preds = xgb.dask.predict(self.dask_client, self.booster, data, **predict_kwargs).compute()

        return preds


OBJECTIVES = {
    "binary:logistic": Tags.BINARY_CLASSIFICATION,
    "reg:logistic": Tags.REGRESSION,
    "reg:squarederror": Tags.REGRESSION,
    "rank:pairwise": Tags.TARGET,
    "rank:ndcg": Tags.TARGET,
    "rank:map": Tags.TARGET,
}


def get_target_tag(objective: str) -> Tags:
    """Get the target tag from the specified objective"""
    try:
        return OBJECTIVES[objective]
    except KeyError as exc:
        target_options_str = str(list(OBJECTIVES.keys()))
        raise ValueError(f"Objective not supported. Must be one of: {target_options_str}") from exc


def get_targets(schema: Schema, target_tag: Tags) -> List[str]:
    """Find target columns from dataset or specified target_column"""
    targets = schema.select_by_tag(Tags.TARGET).select_by_tag(target_tag)

    if len(targets) >= 1:
        return targets.column_names
    raise ValueError(
        f"No target columns in the dataset schema with tags TARGET and {target_tag.name}"
    )


def dataset_to_xy(
    dataset: Dataset,
    target_columns: Union[str, list],
    qid_column: Optional[str],
):
    """Convert Merlin Dataset to XGBoost DMatrix"""
    df = dataset.to_ddf()

    qid = None
    if qid_column:
        df = df.sort_values(qid_column)
        qid = df[qid_column]

    all_target_columns = dataset.schema.select_by_tag(Tags.TARGET).column_names

    # Ignore list-like columns from schema
    list_column_names = [
        col_name
        for col_name, col_schema in dataset.schema.column_schemas.items()
        if col_schema.is_list
    ]

    if list_column_names:
        warnings.warn(f"Ignoring list columns as inputs to XGBoost model: {list_column_names}.")

    X = df.drop(all_target_columns + list_column_names, axis=1)
    y = df[target_columns]

    # Ensure columns are in a consistent order
    X = X[sorted(X.columns)]

    return X, y, qid



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import warnings

import tensorflow as tf
from packaging import version
from tensorflow.python.feature_column import feature_column_v2 as fc

from merlin.core.dispatch import HAS_GPU
from merlin.models.loader.utils import device_mem_size


def configure_tensorflow(memory_allocation=None, device=None):
    total_gpu_mem_mb = device_mem_size(kind="total", cpu=(not HAS_GPU)) / (1024**2)

    if memory_allocation is None:
        memory_allocation = os.environ.get("TF_MEMORY_ALLOCATION", 0.5)

    if float(memory_allocation) < 1:
        memory_allocation = total_gpu_mem_mb * float(memory_allocation)
    memory_allocation = int(memory_allocation)
    assert memory_allocation < total_gpu_mem_mb

    # TODO: what will this look like in any sort
    # of distributed set up?
    if device is None:
        device = int(os.environ.get("TF_VISIBLE_DEVICE", 0))
    tf_devices = tf.config.list_physical_devices("GPU")
    if HAS_GPU and len(tf_devices) == 0:
        raise ImportError("TensorFlow is not configured for GPU")
    if HAS_GPU:
        try:
            tf.config.set_logical_device_configuration(
                tf_devices[device],
                [tf.config.LogicalDeviceConfiguration(memory_limit=memory_allocation)],
            )
        except RuntimeError:
            warnings.warn(
                "TensorFlow runtime already initialized, may not be enough memory for cudf"
            )
        try:
            tf.config.experimental.set_virtual_device_configuration(
                tf_devices[device],
                [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=memory_allocation)],
            )
        except RuntimeError as e:
            # Virtual devices must be set before GPUs have been initialized
            warnings.warn(e)

    # versions using TF earlier than 2.3.0 need to use extension
    # library for dlpack support to avoid memory leak issue
    __TF_DLPACK_STABLE_VERSION = "2.3.0"
    if version.parse(tf.__version__) < version.parse(__TF_DLPACK_STABLE_VERSION):
        try:
            from tfdlpack import from_dlpack
        except ModuleNotFoundError as e:
            message = "If using TensorFlow < 2.3.0, you must install tfdlpack-gpu extension library"
            raise ModuleNotFoundError(message) from e

    else:
        from tensorflow.experimental.dlpack import from_dlpack

    return from_dlpack


def _get_parents(column):
    """
    recursive function for finding the feature columns
    that supply inputs for a given `column`. If there are
    none, returns the column. Uses sets so is not
    deterministic.
    """
    if isinstance(column.parents[0], str):
        return set([column])
    parents = set()
    for parent in column.parents:
        parents |= _get_parents(parent)
    return parents


def get_dataset_schema_from_feature_columns(feature_columns):
    """
    maps from a list of TensorFlow `feature_column`s to
    lists giving the categorical and continuous feature
    names for a dataset. Useful for constructing NVTabular
    Workflows from feature columns
    """
    base_columns = set()
    for column in feature_columns:
        base_columns |= _get_parents(column)

    cat_names, cont_names = [], []
    for column in base_columns:
        if isinstance(column, fc.CategoricalColumn):
            cat_names.append(column.name)
        else:
            cont_names.append(column.name)

    cat_names = sorted(cat_names)
    cont_names = sorted(cont_names)
    return cat_names, cont_names



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from merlin.io import Dataset
from merlin.models.lightfm import LightFM
from merlin.schema import Tags


def test_warp(music_streaming_data: Dataset):
    music_streaming_data.schema = music_streaming_data.schema.remove_by_tag(Tags.TARGET)
    model = LightFM(learning_rate=0.05, loss="warp", epochs=10)
    model.fit(music_streaming_data)

    model.predict(music_streaming_data)
